// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for holding, reading, and writing the calibration key data
// (CKD). Reading, if requested, takes place from another CKD
// file. The data are stored in a NetCDF format.

#pragma once

#include "nc_object.h"
#include "spline_multi.h"

#include <array>
#include <complex>
#include <memory>
#include <netcdf>

namespace gadfit {

class MPIVars;

} // namespace gadfit

namespace spex {

class SPEXSettings;

class CKD
{
private:
    const gadfit::MPIVars& mpi;
    // Read dimensions and data from a NetCDF file up to a target CKD
    // level. The CKD level (see definition below) is then set to that
    // level. The debug argument specifies whether previously stored
    // debug variables should also be read for one or more groups. If
    // a NetCDF group for process i does not exist then skip[i] is set
    // to true, else false. If debug variables were requested but were
    // not found in the NetCDF file, also set debug[i] to false
    // regardless of its previous value. Otherwise, if debug remained
    // true, the write function might attempt to write non-existent
    // variables to an output file.
    auto readCKD(
      const std::string& ckd_file_name,
      const ProcLevel target_level,
      std::array<bool, static_cast<int>(ProcLevel::n_levels)>& debug,
      std::array<bool, static_cast<int>(ProcLevel::n_levels)>& skip,
      const std::array<std::string, static_cast<int>(ProcLevel::n_levels)>&
        alternative_ckd_files) -> void;

public:
    // Attempt to real all CKD contents including the debug
    // information. Nothing is skipped by default. This version can be
    // called by external programs because it doesn't depend on
    // SPEXSettings. Alternative CKD(s) can optionally be used for
    // specified processes. For instance, if
    // !alternative_ckd_file_names[ProcLevel::stray].empty() then the
    // stray light kernels are read from that file and the rest from
    // the main CKD file.
    CKD(const gadfit::MPIVars& mpi,
        const std::string ckd_file_name,
        const std::array<std::string, static_cast<int>(ProcLevel::n_levels)>
          alternative_ckd_file_names = {});
    // Read CKD contents according to SPEXone specific
    // settings. Depending on the target CKD level, not all NetCDF
    // groups are read. Debug data are only read if requested
    // according to the settings (determined by user input
    // parameters). This version can also be called if there is no
    // input CKD file. Then the CKD is initialized using some basic
    // dimension from the settings and constructor returns early.
    CKD(const gadfit::MPIVars& mpi,
        const ProcLevel target_level,
        SPEXSettings& settings);
    // Level of the highest process included in the CKD
    ProcLevel level { ProcLevel::l1a };
    // Number of detector pixels in the spatial direction
    int dim_detector_spat {};
    // Number of detector pixels in the spectral direction
    int dim_detector_spec {};
    // Real dimensions of raw data. Unless
    // [main][detector_spec_reduced] and/or
    // [main][detector_spat_reduced] was given, these are the same as
    // dim_detector_spec and dim_detector_spat.
    int dim_detector_spec_full {};
    int dim_detector_spat_full {};
    // Number of pixels in one unbinned image (dim_detector_spec *
    // dim_detector_spat). If [main][detector_spec_reduced] and/or
    // [main][detector_spat_reduced] is present then npix corresponds
    // to the the reduced image. To get the original number of
    // unbinned pixels use meta->npix.
    int npix {};
    // Number of viewports
    int dim_vp {};
    // Pixel mask is updated during all the detector calibration steps
    // (false = good pixel, true = bad pixel)
    std::vector<bool> pixel_mask {};

    // Dark
    struct
    {
        // Fixed dark signals (independent of integration time)
        SharedArray<double> offset_short {};
        SharedArray<double> offset_long {};
        // Dark current per second of integration time
        SharedArray<double> current {};
        // Sum of least squares of the linear fit per pixel
        SharedArray<double> chi2 {};
        // Uncertainties of dark offset
        SharedArray<double> sigma_beta {};
        // Uncertainties of dark current
        SharedArray<double> sigma_alpha {};
        // Highest exposure times used in the linear fits
        SharedArray<double> last_exposure_times {};
    } dark;

    // Noise
    struct
    {
        // Signal dependent noise term (conversion gain)
        SharedArray<double> g {};
        // Signal independent noise term (scaled dark noise, n)
        SharedArray<double> n {};
        // Standard deviation of g
        SharedArray<double> std_dev_g {};
        // Standard deviation of n
        SharedArray<double> std_dev_n {};
    } noise;

    // Nonlinearity
    static constexpr int nonlin_n_r { 5 }; // r_0 ... r_4
    struct
    {
        // Type of nonlinearity model
        NonlinModel model {};
        // Signal and reference diode rates (step 4)
        std::vector<SharedArray<double>> signal_rates {};
        std::vector<SharedArray<double>> reference_rates {};
        // Reference diode conversion factors (step 5)
        SharedArray<double> f1 {};
        SharedArray<double> f2 {};
        // Maximum signal parameters
        SharedArray<double> m0 {};
        double m1 {};
        double m2 {};
        // Saturated signal values and the corresponding fit (step 8)
        std::vector<double> S_max_data {};
        std::vector<double> S_max_fit {};
        // Detector medians of I_c for the S_max fit
        std::vector<double> I_c_med {};
        // The r and c parameters of Eqs. (72) and (73)
        std::vector<SharedArray<double>> r {};
        SharedArray<double> c {};
        // Copied from SPEXsettings::nonlin.debug_pixels if present
        std::vector<int> debug_pixels {};
        // The scaled linear and measured signals and the
        // corresponding fits for selected pixels
        std::vector<SharedArray<double>> scaled_S_lin {};
        std::vector<SharedArray<double>> scaled_S_meas {};
        std::vector<SharedArray<double>> sigmoid_fits {};
        // Nonlinearity cubic spline coefficients. This is the primary
        // quantity of the nonlinearity CKD.
        SplineMulti spline {};
    } nonlin;
    // Number of B-splines in nonlineary correction fit
    int dim_nonlin_spline {};
    // Number of knots in the B-spline of the nonlinearity fit
    int dim_nonlin_knot {};
    // B-spline order of the nonlinearity correction
    int nonlin_order {};
    // B-spline knots for nonlinearity correction
    std::vector<double> nonlin_knots {};
    // B-spline fit for nonlinearity correction to be used
    std::vector<double> nonlin_fit {};
    // Scaling signal for the abscissa of the nonlinearity fit
    std::vector<double> nonlin_signal_scale {};
    // Separate routine for reading in debug information in case it
    // needs to be called from elsewhere
    auto readNonlinDebugData(const netCDF::NcGroup& grp) -> void;

    // Pixel response nonuniformity (PRNU)
    struct
    {
        SharedArray<double> prnu {};
        // Polynomial fit of the calibration data
        SharedArray<double> sensitivity_fit {};
    } prnu;

    // Stray Light
    struct
    {
        // Number of stray light kernels
        int n_kernels {};
        // Whether to fully or partially unbin images for stray light correction
        bool reduced_kernels {};
        // Number of spatial and spectral elements of the unbinned
        // image. This is determined by the detector dimensions and
        // reduced_kernels.
        int n_spatial {};
        int n_spectral {};
        std::vector<int> kernel_rows {};
        // Number of columns in each kernel
        std::vector<int> kernel_cols {};
        // The FFT array size of each kernel
        std::vector<int> kernel_fft_sizes {};
        // Fourier transforms of the kernels
        std::vector<SharedArray<std::complex<double>>> kernels_fft {};
        // Total internal scattering factor
        SharedArray<double> eta {};
        // Kernel weights
        std::vector<SharedArray<double>> weights {};
        // Boundaries of subimages that must be extracted for the
        // convolutions. The order of coefficients is 'bottom', 'top',
        // 'left', 'right'.
        std::vector<int> edges {};
        // Individual internal scattering factors
        std::vector<SharedArray<double>> etas {};
        int kernel_n_rows_full {};
        int kernel_n_cols_full {};
        std::vector<SharedArray<double>> kernels_real {};
        std::vector<double> all_kernels {};
    } stray;

    // Field of view (FOV)
    struct
    {
        // Number of spectra per viewport
        std::vector<int> n_act_vp {};
        // Number of spectra in one swath (sum over all n_act_vp)
        int n_act {};
        // Starting FOV index of the current viewport
        std::vector<int> start_vp {};
        // List of across track angles determined by calibration
        // measurements and user input. These are used in later CKD
        // generation steps and in the L1B processor.
        std::vector<double> act_angles {};
        // Fractional spatial (row) indices of spectra. This is the
        // primary quantity of the FOV CKD.
        std::array<std::vector<SharedArray<double>>, dims::pol> spat_indices {};
        // First pixel (down pixel) involved in interpolation of L1B
        // spectra, derived from spat_indices
        std::array<std::vector<SharedArray<int>>, dims::pol> pix1 {};
        // Second (up) pixel involved in interpolation of L1B spectra
        // derived from spat_indices
        std::array<std::vector<SharedArray<int>>, dims::pol> pix2 {};
        // Weight factor of the first pixel in this interpolation
        std::array<std::vector<SharedArray<double>>, dims::pol> weight {};
        // Size of spectra of all ACT angles
        std::vector<int> dims_spec {};
        // Number of calibration measurements that are included in the
        // FOV CKD generation. Those that fall outside the visible
        // area of the detector are not included. This variable is
        // updated at the end of each batch (viewport) of
        // measurements. Within the processBatch function it is the
        // number of visible calibration measurements over all
        // previous viewports.
        int n_l1a_total {};
        // Total number of detector columns to be processed. This can
        // differ from the detector dimension if some columns from
        // left and/or right side of the detector are excluded.
        int n_columns {};
        // Positions of peaks as determined by Gaussian fits along
        // columns. The dimensions are (dims::pol, n_l1a_total,
        // dim_detector_spec).
        std::array<std::vector<SharedArray<double>>, dims::pol> peak_pos {};
        // B-spline fits to the peaks positions along the spectral
        // dimension. The objective is to smooth out noise in the
        // spatial dimension.
        std::array<std::vector<SharedArray<double>>, dims::pol> peak_pos_b {};
    } fov;

    // Swath vectors
    // Pointing vector per FOV in satellite coordinates
    std::vector<double> swath_swathvectors {};
    // Normals of planes fitted through swath vectors of one viewport
    std::vector<double> swath_vectorplane_normals {};

    // Wavelength
    // Wavelengths of spectra corresponding to each FOV and
    // polarization state
    std::array<std::vector<SharedArray<double>>, dims::pol> wave_full {};
    // Common wavelength grid for each pair of S+ and S- spectra
    std::vector<SharedArray<double>> wave_common {};

    // Radiometric
    // Radiometric calibration factor for extracted spectra
    std::array<std::vector<SharedArray<double>>, dims::pol> rad_spectra {};

    // Polarimetric
    struct
    {
        // Mueller matrix elements for q
        std::vector<SharedArray<double>> m_q {};
        // Mueller matrix elements for u
        std::vector<SharedArray<double>> m_u {};
        // Mueller matrix elements for telescope polarization
        std::vector<SharedArray<double>> m_t {};
        // Polarimetric efficiency
        std::vector<SharedArray<double>> eps {};
        // Polarimetric retardance
        std::vector<SharedArray<double>> delta {};
    } pol;

    // Write CKD into a NetCDF output file. It is up to the caller to
    // set the CKD level before and after calling this function.
    auto write(const SPEXSettings& settings) -> void;

    ~CKD() = default;
};

} // namespace spex
