// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1c.h"

#include "constants.h"
#include "cubic_spline.h"
#include "l1b.h"
#include "spex_math.h"

#include <numeric>
#include <spdlog/spdlog.h>

namespace spex {

// Scale factor for some of the geolocation data variables
constexpr float geo_scale_factor { 100.0f };

// Main dimensions used in this module
struct Dims
{
    // Number of rows in the list of L1C latitudes (nc variable
    // bins_across_track)
    int n_l1c_rows {};
    // Number of columns in the list of L1C latitudes (nc variable
    // bins_across_track)
    int n_l1c_cols {};
    // Number of rows and columns in the dense L1C grid
    int n_l1c_rows_dbl {};
    int n_l1c_cols_dbl {};
    // Number L1B rows and columns
    int n_l1b_rows {};
    int n_l1b_cols {};
    // Lowest row in the list of L1B latitudes used for interpolation
    int min_l1b_row { math::idx_fill_value };
    // Highest L1b row used for interpolation
    int max_l1b_row { math::idx_fill_value };
    std::vector<double> row_indices {};
    std::vector<double> col_indices {};
    // L1B columns indices interpolated along the L1C rows
    std::vector<double> interp_cols {};
    // For a L1C given row, number of columns on the left that should
    // be set to fill values because they are outside the L1B grid.
    std::vector<int> fill_cols_left {};
    // Index of the first L1C column on the right that is outside the
    // L1B swath and should set to fill values.
    std::vector<int> fill_cols_right {};
};

// Unbin and interpolate data from an MxN grid to a 2Mx2N grid
static auto expandGrid(const int n_rows,
                       const int n_cols,
                       std::vector<float>& data) -> void
{
    constexpr int bin { 2 };
    // First interpolate across rows
    std::vector<double> data_semi_dense(bin * data.size());
    std::vector<double> x_values(n_rows);
    std::vector<double> y_values(n_rows);
    for (int j {}; j < n_cols; ++j) {
        for (int i {}; i < n_rows; ++i) {
            x_values[i] = static_cast<double>(i);
            y_values[i] = static_cast<double>(data[i * n_cols + j]);
        }
        spex::CubicSpline spline { x_values, y_values };
        for (int i {}; i < bin * n_rows; ++i) {
            // Index 0 becomes -0.25, 0.25 after unbinning. When
            // working with the dense we move in steps of 0.5 in terms
            // of the original grid.
            const double arg_x { -0.25 + i * 0.5 };
            data_semi_dense[i * n_cols + j] = spline.eval(arg_x);
        }
    }
    // Next interpolate across columns
    std::vector<float> data_dense(bin * bin * data.size());
    x_values.resize(n_cols);
    y_values.resize(n_cols);
    for (int i {}; i < bin * n_rows; ++i) {
        for (int j {}; j < n_cols; ++j) {
            x_values[j] = j;
            y_values[j] = data_semi_dense[i * n_cols + j];
        }
        spex::CubicSpline spline { x_values, y_values };
        for (int j {}; j < bin * n_cols; ++j) {
            const double arg_y { -0.25 + j * 0.5 };
            data_dense[i * bin * n_cols + j] =
              static_cast<float>(spline.eval(arg_y));
        }
    }
    std::swap(data, data_dense);
}

static auto crossProduct(const std::span<double> a,
                         const std::span<double> b,
                         std::array<double, 3>& c) -> void
{
    c[0] = a[1] * b[2] - a[2] * b[1];
    c[1] = a[2] * b[0] - a[0] * b[2];
    c[2] = a[0] * b[1] - a[1] * b[0];
}

static auto vectorNorm(const std::array<double, dims::vec>& v) -> double
{
    return std::sqrt(std::accumulate(
      v.cbegin(), v.cend(), 0.0, [](const auto& s, const auto& x) {
          return s + x * x;
      }));
}

// Read in the L1C latitudes/longitudes, create a dense L1C, and
// compute the L1C normal vectors.
static auto prepareL1CGrid(const SPEXSettings& settings,
                           Dims& dims,
                           std::vector<float>& l1c_latitudes,
                           std::vector<float>& l1c_longitudes,
                           std::vector<double>& normal_vectors_r,
                           std::vector<double>& normal_vectors_c) -> void
{
    NcObject nc_l1c { netCDF::NcFile::read, settings.l1c.l1c_in };
    dims.n_l1c_rows =
      static_cast<int>(nc_l1c.getDim("bins_along_track").getSize());
    dims.n_l1c_cols =
      static_cast<int>(nc_l1c.getDim("bins_across_track").getSize());
    dims.n_l1c_rows_dbl = 2 * dims.n_l1c_rows;
    dims.n_l1c_cols_dbl = 2 * dims.n_l1c_cols;
    l1c_latitudes.resize(dims.n_l1c_rows * dims.n_l1c_cols);
    l1c_longitudes.resize(dims.n_l1c_rows * dims.n_l1c_cols);
    const auto geolocation_data { nc_l1c.getGroup("geolocation_data") };
    geolocation_data.getVar("latitude").getVar(l1c_latitudes.data());
    for (auto& val : l1c_latitudes) {
        val = static_cast<float>(static_cast<double>(val) * math::deg_to_rad);
    }
    geolocation_data.getVar("longitude").getVar(l1c_longitudes.data());
    for (auto& val : l1c_longitudes) {
        val = static_cast<float>(static_cast<double>(val) * math::deg_to_rad);
    }
    expandGrid(dims.n_l1c_rows, dims.n_l1c_cols, l1c_latitudes);
    expandGrid(dims.n_l1c_rows, dims.n_l1c_cols, l1c_longitudes);
    // Next compute the L1C unit vectors from latitude/longitude
    normal_vectors_r.resize(dims.n_l1c_rows_dbl * dims::vec);
    normal_vectors_c.resize(dims.n_l1c_rows_dbl * dims::vec);
    //     ( cos(lon)cos(lat) )
    // G = ( sin(lon)cos(lat) )
    //     ( sin(lat)         )
    std::vector<double> l1c_unit_vectors(dims.n_l1c_rows_dbl
                                         * dims.n_l1c_cols_dbl * dims::vec);
    for (int i {}; i < dims.n_l1c_rows_dbl * dims.n_l1c_cols_dbl; ++i) {
        l1c_unit_vectors[i * dims::vec] =
          std::cos(l1c_longitudes[i]) * std::cos(l1c_latitudes[i]);
        l1c_unit_vectors[i * dims::vec + 1] =
          std::sin(l1c_longitudes[i]) * std::cos(l1c_latitudes[i]);
        l1c_unit_vectors[i * dims::vec + 2] = std::sin(l1c_latitudes[i]);
    }

    // Compute the L1C row and column normal vectors
    std::array<double, dims::vec> cross {}; // cross product
    const int mid_col { dims.n_l1c_cols_dbl / 2 };
    for (int i_row {}; i_row < dims.n_l1c_rows_dbl; ++i_row) {
        crossProduct(
          // Last column of the row
          std::span(&l1c_unit_vectors[((i_row + 1) * dims.n_l1c_cols_dbl - 1)
                                      * dims::vec],
                    dims::vec),
          // First column of the row
          std::span(&l1c_unit_vectors[i_row * dims.n_l1c_cols_dbl * dims::vec],
                    dims::vec),
          cross);
        const double norm { vectorNorm(cross) };
        for (int i_vec {}; i_vec < dims::vec; ++i_vec) {
            normal_vectors_r[i_row * dims::vec + i_vec] = cross[i_vec] / norm;
        }
        // Because there is an even number of L1C columns, the middle
        // vector is defined as the average of the middle two.
        std::array<double, dims::vec> mid_vector {};
        for (int i_vec {}; i_vec < dims::vec; ++i_vec) {
            mid_vector[i_vec] =
              (l1c_unit_vectors[(i_row * dims.n_l1c_cols_dbl + mid_col)
                                  * dims::vec
                                + i_vec]
               + l1c_unit_vectors[(i_row * dims.n_l1c_cols_dbl + mid_col + 1)
                                    * dims::vec
                                  + i_vec])
              / 2;
        }
        const double norm_mid { vectorNorm(mid_vector) };
        for (auto& val : mid_vector) {
            val /= norm_mid;
        }
        crossProduct(std::span(mid_vector.data(), dims::vec),
                     std::span(&normal_vectors_r[i_row * dims::vec], dims::vec),
                     cross);
        for (int i_vec {}; i_vec < dims::vec; ++i_vec) {
            normal_vectors_c[i_row * dims::vec + i_vec] = cross[i_vec];
        }
    }
}

static auto dotProduct(const std::span<double> a, const std::span<double> b)
  -> double
{
    double sum {};
    for (int i_vec {}; i_vec < dims::vec; ++i_vec) {
        sum += a[i_vec] * b[i_vec];
    }
    return sum;
}

// Assign a level 1C row and column to each L1B grid point
static auto genL1CCoordinates(const SPEXSettings& settings,
                              const std::vector<double>& l1b_lat_vp,
                              const std::vector<double>& l1b_lon_vp,
                              const std::vector<float>& l1c_latitudes,
                              const std::vector<float>& l1c_longitudes,
                              std::vector<double>& normal_vectors_r,
                              std::vector<double>& normal_vectors_c,
                              Dims& dims) -> void
{
    // Compute the L1B unit vectors (same formula as above)
    std::vector<double> l1b_unit_vectors(dims.n_l1b_rows * dims.n_l1b_cols
                                         * dims::vec);
    std::vector<double> l1c_unit_vectors(dims.n_l1c_rows_dbl
                                         * dims.n_l1c_cols_dbl * dims::vec);
    for (int i {}; i < dims.n_l1b_rows * dims.n_l1b_cols; ++i) {
        l1b_unit_vectors[i * dims::vec] =
          std::cos(l1b_lon_vp[i]) * std::cos(l1b_lat_vp[i]);
        l1b_unit_vectors[i * dims::vec + 1] =
          std::sin(l1b_lon_vp[i]) * std::cos(l1b_lat_vp[i]);
        l1b_unit_vectors[i * dims::vec + 2] = std::sin(l1b_lat_vp[i]);
    }

    // Compute the target L1C row and column indices
    std::vector<int> l1c_rows(dims.n_l1b_rows * dims.n_l1b_cols,
                              math::idx_fill_value);
    std::vector<int> l1c_cols(dims.n_l1b_rows * dims.n_l1b_cols,
                              math::idx_fill_value);
    // Divide the resolution by two because we are working with the
    // dense L1C grid. Delta is the threshold for the dot product that
    // determines if we have found the correct L1C row.
    const double delta { settings.l1c.l1c_resolution / 2
                         / settings.l1c.semi_major_axis / 2 };
    const int mid_col { dims.n_l1c_cols_dbl / 2 };
    for (int i_l1b {}; i_l1b < dims.n_l1b_rows * dims.n_l1b_cols; ++i_l1b) {
        for (int i_l1c {}; i_l1c < dims.n_l1c_rows_dbl; ++i_l1c) {
            const double angle_r { dotProduct(
              std::span(&l1b_unit_vectors[i_l1b * dims::vec], dims::vec),
              std::span(&normal_vectors_r[i_l1c * dims::vec], dims::vec)) };
            // Include a small margin to account for numerical
            // precision. It could be that if a point sits exactly in
            // between two L1C rows it not associated with either of
            // them.
            constexpr double prec_margin { 2e-6 };
            if (std::abs(angle_r) <= delta + prec_margin) {
                l1c_rows[i_l1b] = (angle_r < 0.0) ? i_l1c : i_l1c - 1;
                if (l1c_rows[i_l1b] >= dims.n_l1c_rows_dbl - 1) {
                    l1c_rows[i_l1b] = math::idx_fill_value;
                }
                const double angle_c { dotProduct(
                  std::span(&l1b_unit_vectors[i_l1b * dims::vec], dims::vec),
                  std::span(&normal_vectors_c[i_l1c * dims::vec], dims::vec)) };
                l1c_cols[i_l1b] =
                  static_cast<int>(mid_col + 0.5 + angle_c / (2 * delta));
                break;
            }
        }
    }

    // Find usable L1B range
    for (int i_row {}; i_row < dims.n_l1b_rows; ++i_row) {
        bool row_clean { true };
        for (int i_col {}; i_col < dims.n_l1b_cols; ++i_col) {
            if (l1c_rows[i_row * dims.n_l1b_cols + i_col]
                == math::idx_fill_value) {
                row_clean = false;
            }
        }
        if (row_clean) {
            if (dims.min_l1b_row == math::idx_fill_value) {
                dims.min_l1b_row = i_row;
            }
            dims.max_l1b_row = i_row + 1;
        }
    }
    // Extend the range a little bit in order to make the
    // interpolation more accurate. For this we copy the L1C row and
    // column indices to the rows that outside the L1C range. That is
    // strictly speaking not correct but then the relative coordinates
    // would assume large negative or positive values values which
    // makes it okay. Basically, we are linearly interpolating the L1C
    // row indices outside its range.
    constexpr int n_extra_l1b_rows { 2 };
    for (int i_row { dims.min_l1b_row - n_extra_l1b_rows };
         i_row < dims.min_l1b_row;
         ++i_row) {
        for (int i_col {}; i_col < dims.n_l1b_cols; ++i_col) {
            if (i_row >= 0) {
                l1c_rows[i_row * dims.n_l1b_cols + i_col] =
                  l1c_rows[dims.min_l1b_row * dims.n_l1b_cols + i_col];
                l1c_cols[i_row * dims.n_l1b_cols + i_col] =
                  l1c_cols[dims.min_l1b_row * dims.n_l1b_cols + i_col];
            }
        }
    }
    dims.min_l1b_row = std::max(0, dims.min_l1b_row - n_extra_l1b_rows);
    for (int i_row { dims.max_l1b_row };
         i_row < dims.max_l1b_row + n_extra_l1b_rows;
         ++i_row) {
        for (int i_col {}; i_col < dims.n_l1b_cols; ++i_col) {
            if (i_row < dims.n_l1b_rows) {
                l1c_rows[i_row * dims.n_l1b_cols + i_col] =
                  l1c_rows[(dims.max_l1b_row - 1) * dims.n_l1b_cols + i_col];
                l1c_cols[i_row * dims.n_l1b_cols + i_col] =
                  l1c_cols[(dims.max_l1b_row - 1) * dims.n_l1b_cols + i_col];
            }
        }
    }
    dims.max_l1b_row =
      std::min(dims.n_l1b_rows, dims.max_l1b_row + n_extra_l1b_rows);
    spdlog::debug("Using L1B rows {}...{} for the interpolation",
                  dims.min_l1b_row,
                  dims.max_l1b_row);

    // Make sure there are no bad L1C row or column indices within the
    // usable L1B row range.
    for (int i_row_l1b { dims.min_l1b_row }; i_row_l1b < dims.max_l1b_row;
         ++i_row_l1b) {
        for (int i_col_l1b {}; i_col_l1b < dims.n_l1b_cols; ++i_col_l1b) {
            const int idx { i_row_l1b * dims.n_l1b_cols + i_col_l1b };
            if (l1c_rows[idx] == math::idx_fill_value
                || l1c_cols[idx] == math::idx_fill_value) {
                throw std::runtime_error {
                    "L1B point (" + std::to_string(i_row_l1b) + ", "
                    + std::to_string(i_col_l1b)
                    + ") has a bad L1C row/column index: "
                    + std::to_string(l1c_rows[idx]) + ", "
                    + std::to_string(l1c_cols[idx])
                };
            }
        }
    }

    // Compute the relative coordinates of an input point within an
    // L1C cell and then the full coordinates in terms of fractional
    // L1C row and column indices.
    dims.row_indices.resize(dims.n_l1b_rows * dims.n_l1b_cols);
    dims.col_indices.resize(dims.n_l1b_rows * dims.n_l1b_cols);
    for (int i_row_l1b { dims.min_l1b_row }; i_row_l1b < dims.max_l1b_row;
         ++i_row_l1b) {
        for (int i_col_l1b {}; i_col_l1b < dims.n_l1b_cols; ++i_col_l1b) {
            const int l1c_row {
                l1c_rows[i_row_l1b * dims.n_l1b_cols + i_col_l1b]
            };
            const int l1c_col {
                l1c_cols[i_row_l1b * dims.n_l1b_cols + i_col_l1b]
            };
            // Input location
            std::array<double, 2> p_x {
                l1b_lon_vp[i_row_l1b * dims.n_l1b_cols + i_col_l1b],
                l1b_lat_vp[i_row_l1b * dims.n_l1b_cols + i_col_l1b]
            };
            // Identify the three corners of the target L1C cell. The
            // fourth corner p11 is not needed for anything and is
            // thus omitted here.
            std::array<double, 2> p_00 {
                l1c_longitudes[l1c_row * dims.n_l1c_cols_dbl + l1c_col],
                l1c_latitudes[l1c_row * dims.n_l1c_cols_dbl + l1c_col]
            };
            std::array<double, 2> p_01 {
                l1c_longitudes[l1c_row * dims.n_l1c_cols_dbl + l1c_col + 1],
                l1c_latitudes[l1c_row * dims.n_l1c_cols_dbl + l1c_col + 1]
            };
            std::array<double, 2> p_10 {
                l1c_longitudes[(l1c_row + 1) * dims.n_l1c_cols_dbl + l1c_col],
                l1c_latitudes[(l1c_row + 1) * dims.n_l1c_cols_dbl + l1c_col]
            };

            // Shift the local coordinate system origin to zero. The
            // origin is defined to be p_00.
            p_x[0] -= p_00[0];
            p_x[1] -= p_00[1];
            p_01[0] -= p_00[0];
            p_01[1] -= p_00[1];
            p_10[0] -= p_00[0];
            p_10[1] -= p_00[1];
            // p_00 becomes 0 and is thus omitted wherever it needs to be used

            // Rotate along the x-axis such that the latitude of the
            // right bottom cell point is zero.
            const double theta { -std::atan(p_01[1] / p_01[0]) };
            std::array<double, 4> rot_mat { std::cos(theta),
                                            -std::sin(theta),
                                            std::sin(theta),
                                            std::cos(theta) };
            // Return Rx where R is the rotation matrix
            const auto multRotMat { [&rot_mat](const std::array<double, 2>& x) {
                return std::array<double, 2> {
                    rot_mat[0] * x[0] + rot_mat[1] * x[1],
                    rot_mat[2] * x[0] + rot_mat[3] * x[1]
                };
            } };
            p_x = multRotMat(p_x);
            // p_00 is unaffected by this rotation
            p_01 = multRotMat(p_01);
            p_10 = multRotMat(p_10);

            // Undo skew
            const double alpha { 90.0 * math::deg_to_rad
                                 - std::atan(p_10[0] / p_10[1]) };
            rot_mat = { 1.0,
                        -1.0 / std::tan(alpha),
                        0.0,
                        std::abs(1.0 / std::sin(alpha)) };
            p_x = multRotMat(p_x);
            // p_00 is unaffected by this rotation
            p_01 = multRotMat(p_01);
            p_10 = multRotMat(p_10);

            // Normalize the relative coordinates to be in the range 0..1
            const double Dx { std::abs(p_01[0]) };
            const double Dy { std::abs(p_10[1]) };
            p_x = { p_x[0] / Dx, p_x[1] / Dy };

            // Add the relative cell coordinates to the row and cell
            // numbers to get the final L1B coordinates in terms of
            // the fractional L1C row and column numbers.
            dims.row_indices[i_row_l1b * dims.n_l1b_cols + i_col_l1b] =
              l1c_rows[i_row_l1b * dims.n_l1b_cols + i_col_l1b] + p_x[1];
            dims.col_indices[i_row_l1b * dims.n_l1b_cols + i_col_l1b] =
              l1c_cols[i_row_l1b * dims.n_l1b_cols + i_col_l1b] + p_x[0];

            // Notify if on or both of the relative coordinates are too negative
            constexpr double err_margin { 0.03 };
            if (p_x[0] < 0.0 - err_margin || p_x[0] > 1.0 + err_margin
                || p_x[1] < 0.0 - err_margin || p_x[1] > 1.0 + err_margin) {
                spdlog::debug("Relative coordinates of L1B point ({}, {}) are "
                              "{:6.3f}, {:6.3f}",
                              i_row_l1b,
                              i_col_l1b,
                              p_x[0],
                              p_x[1]);
            }
        }
    }
}

// 2D interpolate a quantity stored in l1b_buf and 2x2 bin the result
// and store it in l1c_buf.
static auto interpolateAndBin(Dims& dims,
                              const std::vector<double>& l1b_buf,
                              std::vector<double>& l1c_buf) -> void
{
    // First interpolate across the L1B rows
    const int n_rows { dims.max_l1b_row - dims.min_l1b_row };
    // L1B data after interpolation across rows
    std::vector<double> data_semi(dims.n_l1c_rows_dbl * dims.n_l1b_cols);
    // x-values for the first interpolation are the fractional L1C row indices
    std::vector<double> x_values(n_rows);
    // y-values are the L1B data or the fractional L1C column indices
    // if the latter has not been computed yet.
    std::vector<double> y_values(n_rows);
    // The new column indices need to be computed once per viewport
    if (dims.interp_cols.empty()) {
        dims.interp_cols.resize(dims.n_l1c_rows_dbl * dims.n_l1b_cols);
        for (int j {}; j < dims.n_l1b_cols; ++j) {
            for (int i {}; i < n_rows; ++i) {
                const int idx { (dims.min_l1b_row + i) * dims.n_l1b_cols + j };
                x_values[i] = dims.row_indices[idx];
                y_values[i] = dims.col_indices[idx];
            }
            spex::CubicSpline spline { x_values, y_values };
            for (int i {}; i < dims.n_l1c_rows_dbl; ++i) {
                dims.interp_cols[i * dims.n_l1b_cols + j] =
                  spline.eval(static_cast<double>(i));
            }
        }
    }
    // L1B columns indices interpolated along the L1C rows
    for (int j {}; j < dims.n_l1b_cols; ++j) {
        for (int i {}; i < n_rows; ++i) {
            const int idx { (dims.min_l1b_row + i) * dims.n_l1b_cols + j };
            x_values[i] = dims.row_indices[idx];
            y_values[i] = l1b_buf[idx];
        }
        spex::CubicSpline spline { x_values, y_values };
        for (int i {}; i < dims.n_l1c_rows_dbl; ++i) {
            data_semi[i * dims.n_l1b_cols + j] =
              spline.eval(static_cast<double>(i));
        }
    }
    // Determine which L1C columns are outside the L1B swath. This
    // needs to be done once per viewport
    if (dims.fill_cols_left.empty()) {
        dims.fill_cols_left.resize(dims.row_indices.size(),
                                   math::idx_fill_value);
        dims.fill_cols_right.resize(dims.row_indices.size(),
                                    math::idx_fill_value);
        for (int i {}; i < dims.n_l1c_rows_dbl; ++i) {
            const double l1b_col_first {
                dims.interp_cols[i * dims.n_l1b_cols]
            };
            const double l1b_col_last {
                dims.interp_cols[(i + 1) * dims.n_l1b_cols - 1]
            };
            for (int j {}; j < dims.n_l1c_cols_dbl; ++j) {
                if (j < l1b_col_first) {
                    dims.fill_cols_left[i] = j + 1;
                }
                if (j > l1b_col_last) {
                    dims.fill_cols_right[i] = j;
                    // Capture the first column on the right that is
                    // outside the L1B range. Do not process further.
                    break;
                }
            }
        }
    }
    // Next interpolate along the L1C rows. The x-values are now the
    // interpolated L1C column indices and y-values are the L1B data
    // interpolated onto the semi-dense grid.
    x_values.resize(dims.n_l1b_cols);
    y_values.resize(dims.n_l1b_cols);
    std::vector<double> l1c_buf_unbin(dims.n_l1c_rows_dbl
                                      * dims.n_l1c_cols_dbl);
    for (int i {}; i < dims.n_l1c_rows_dbl; ++i) {
        for (int j {}; j < dims.n_l1b_cols; ++j) {
            x_values[j] = dims.interp_cols[i * dims.n_l1b_cols + j];
            y_values[j] = data_semi[i * dims.n_l1b_cols + j];
        }
        spex::CubicSpline spline { x_values, y_values };
        for (int j {}; j < dims.fill_cols_left[i]; ++j) {
            l1c_buf_unbin[i * dims.n_l1c_cols_dbl + j] = NC_FILL_DOUBLE;
        }
        for (int j { dims.fill_cols_left[i] }; j < dims.fill_cols_right[i];
             ++j) {
            l1c_buf_unbin[i * dims.n_l1c_cols_dbl + j] =
              spline.eval(static_cast<double>(j));
        }
        for (int j { dims.fill_cols_right[i] }; j < dims.n_l1c_cols_dbl; ++j) {
            l1c_buf_unbin[i * dims.n_l1c_cols_dbl + j] = NC_FILL_DOUBLE;
        }
    }
    // Finally, bin the data and store in the L1C buffer
    constexpr int bin { 2 };
    std::fill(l1c_buf.begin(), l1c_buf.end(), 0.0);
    for (int i {}; i < dims.n_l1c_rows; ++i) {
        for (int j {}; j < dims.n_l1c_cols; ++j) {
            for (int m {}; m < bin; ++m) {
                for (int n {}; n < bin; ++n) {
                    l1c_buf[i * dims.n_l1c_cols + j] +=
                      l1c_buf_unbin[(bin * i + m) * dims.n_l1c_cols_dbl
                                    + bin * j + n];
                }
            }
        }
    }
    const double norm { 1.0 / bin / bin };
    for (auto& val : l1c_buf) {
        val *= norm;
        if (std::abs(val) > std::sqrt(NC_FILL_DOUBLE)) {
            val = NC_FILL_DOUBLE;
        }
    }
}

// Convert and a float array to an integer array while multiplying
// each element with a scale factor to preserve accuracy.
static auto floatToShort(const SharedArray<float>& data_in,
                         std::vector<int16_t>& data_out) -> void
{
    data_out.resize(data_in.size());
    for (int i {}; i < static_cast<int>(data_in.size()); ++i) {
        data_out[i] = static_cast<int16_t>(geo_scale_factor * data_in[i]);
    }
}

auto L1C::processInit() -> void
{
    // Read in and expand the L1C target grid
    Dims dims {};
    std::vector<float> l1c_latitudes {};
    std::vector<float> l1c_longitudes {};
    std::vector<double> normal_vectors_r {};
    std::vector<double> normal_vectors_c {};
    prepareL1CGrid(settings,
                   dims,
                   l1c_latitudes,
                   l1c_longitudes,
                   normal_vectors_r,
                   normal_vectors_c);

    // L1B dimensions
    NcObject nc_l1b { netCDF::NcFile::read, settings.l1c.l1b_in };
    const int number_of_views {
        static_cast<int>(nc_l1b.getDim("number_of_views").getSize()),
    };
    dims.n_l1b_rows =
      static_cast<int>(nc_l1b.getDim("bins_along_track").getSize());
    const int spatial_samples_per_image {
        static_cast<int>(nc_l1b.getDim("spatial_samples_per_image").getSize()),
    };
    const int intensity_bands_per_view {
        static_cast<int>(nc_l1b.getDim("intensity_bands_per_view").getSize()),
    };
    const int polarization_bands_per_view {
        static_cast<int>(
          nc_l1b.getDim("polarization_bands_per_view").getSize()),
    };

    // L1B sensor view bands
    const auto l1b_sensor_view_bands { nc_l1b.getGroup("SENSOR_VIEW_BANDS") };
    std::vector<uint8_t> viewport_index(spatial_samples_per_image);
    l1b_sensor_view_bands.getVar("viewport_index")
      .getVar(viewport_index.data());
    std::vector<float> intensity_wavelengths(number_of_views
                                             * intensity_bands_per_view);
    l1b_sensor_view_bands.getVar("intensity_wavelengths")
      .getVar(intensity_wavelengths.data());
    std::vector<float> polarization_wavelengths(number_of_views
                                                * polarization_bands_per_view);
    l1b_sensor_view_bands.getVar("polarization_wavelengths")
      .getVar(polarization_wavelengths.data());

    std::vector<double> image_time(dims.n_l1b_rows);
    nc_l1b.getGroup("BIN_ATTRIBUTES")
      .getVar("image_time")
      .getVar(image_time.data());

    // L1B geolocation data
    const auto geolocation_data { nc_l1b.getGroup("GEOLOCATION_DATA") };
    std::vector<float> l1b_latitudes(dims.n_l1b_rows
                                     * spatial_samples_per_image);
    geolocation_data.getVar("latitude").getVar(l1b_latitudes.data());
    for (auto& val : l1b_latitudes) {
        val = static_cast<float>(static_cast<double>(val) * math::deg_to_rad);
    }
    std::vector<float> l1b_longitudes(dims.n_l1b_rows
                                      * spatial_samples_per_image);
    geolocation_data.getVar("longitude").getVar(l1b_longitudes.data());
    for (auto& val : l1b_longitudes) {
        val = static_cast<float>(static_cast<double>(val) * math::deg_to_rad);
    }

    // L1B observation (physical) data
    const auto l1b_observation_data { nc_l1b.getGroup("OBSERVATION_DATA") };
    const auto l1b_geolocation_data { nc_l1b.getGroup("GEOLOCATION_DATA") };
    const int l1b_intensity_size { dims.n_l1b_rows * spatial_samples_per_image
                                   * intensity_bands_per_view };
    const int l1b_pol_size { dims.n_l1b_rows * spatial_samples_per_image
                             * polarization_bands_per_view };
    SharedArray<float> l1b_I { true, mpi, l1b_intensity_size };
    SharedArray<float> l1b_I_noise { true, mpi, l1b_intensity_size };
    SharedArray<float> l1b_I_polsample { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_I_polsample_noise { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_q { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_q_noise { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_u { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_u_noise { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_dolp { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_dolp_noise { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_aolp { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_aolp_noise { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_altitude { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_sensor_zenith { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_sensor_azimuth { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_solar_zenith { true, mpi, l1b_pol_size };
    SharedArray<float> l1b_solar_azimuth { true, mpi, l1b_pol_size };
    if (mpi.rank == 0) {
        l1b_observation_data.getVar("I").getVar(l1b_I.data());
        l1b_observation_data.getVar("I_noise").getVar(l1b_I_noise.data());
        l1b_observation_data.getVar("I_polsample")
          .getVar(l1b_I_polsample.data());
        l1b_observation_data.getVar("I_polsample_noise")
          .getVar(l1b_I_polsample_noise.data());
        l1b_observation_data.getVar("q").getVar(l1b_q.data());
        l1b_observation_data.getVar("q_noise").getVar(l1b_q_noise.data());
        l1b_observation_data.getVar("u").getVar(l1b_u.data());
        l1b_observation_data.getVar("u_noise").getVar(l1b_u_noise.data());
        l1b_observation_data.getVar("DoLP").getVar(l1b_dolp.data());
        l1b_observation_data.getVar("DoLP_noise").getVar(l1b_dolp_noise.data());
        l1b_observation_data.getVar("AoLP").getVar(l1b_aolp.data());
        l1b_observation_data.getVar("AoLP_noise").getVar(l1b_aolp_noise.data());
        l1b_geolocation_data.getVar("altitude").getVar(l1b_altitude.data());
        l1b_geolocation_data.getVar("sensor_zenith")
          .getVar(l1b_sensor_zenith.data());
        l1b_geolocation_data.getVar("sensor_azimuth")
          .getVar(l1b_sensor_azimuth.data());
        l1b_geolocation_data.getVar("solar_zenith")
          .getVar(l1b_solar_zenith.data());
        l1b_geolocation_data.getVar("solar_azimuth")
          .getVar(l1b_solar_azimuth.data());
    }

    // L1C observation data
    const int l1c_intensity_size { dims.n_l1c_rows * dims.n_l1c_cols
                                   * number_of_views
                                   * intensity_bands_per_view };
    const int l1c_pol_size { dims.n_l1c_rows * dims.n_l1c_cols * number_of_views
                             * polarization_bands_per_view };
    const int l1c_latlon_size { dims.n_l1c_rows * dims.n_l1c_cols };
    // Like lat/lon but with angular dependence
    const int l1c_ang_size { dims.n_l1c_rows * dims.n_l1c_cols
                             * number_of_views };
    SharedArray<float> l1c_I { true, mpi, l1c_intensity_size };
    SharedArray<float> l1c_I_noise { true, mpi, l1c_intensity_size };
    SharedArray<float> l1c_I_polsample { true, mpi, l1c_pol_size };
    SharedArray<float> l1c_I_polsample_noise { true, mpi, l1c_pol_size };
    SharedArray<float> l1c_q { true, mpi, l1c_pol_size };
    SharedArray<float> l1c_q_noise { true, mpi, l1c_pol_size };
    SharedArray<float> l1c_u { true, mpi, l1c_pol_size };
    SharedArray<float> l1c_u_noise { true, mpi, l1c_pol_size };
    SharedArray<float> l1c_dolp { true, mpi, l1c_pol_size };
    SharedArray<float> l1c_dolp_noise { true, mpi, l1c_pol_size };
    SharedArray<float> l1c_aolp { true, mpi, l1c_pol_size };
    SharedArray<float> l1c_aolp_noise { true, mpi, l1c_pol_size };
    SharedArray<float> l1c_altitude { true, mpi, l1c_latlon_size };
    SharedArray<float> l1c_sza { true, mpi, l1c_ang_size };
    SharedArray<float> l1c_saa { true, mpi, l1c_ang_size };
    SharedArray<float> l1c_vza { true, mpi, l1c_ang_size };
    SharedArray<float> l1c_vaa { true, mpi, l1c_ang_size };

    for (int i_vp {}; i_vp < number_of_views; ++i_vp) {
        spdlog::info("Processing viewport {}", i_vp);
        // Quantities that need to be reset and recomputed for each viewport
        dims.min_l1b_row = math::idx_fill_value;
        dims.max_l1b_row = math::idx_fill_value;
        dims.fill_cols_left = std::vector<int> {};
        dims.fill_cols_right = std::vector<int> {};
        dims.interp_cols = std::vector<double> {};
        // Number of across track bins of the current viewport
        dims.n_l1b_cols =
          static_cast<int>(std::ranges::count(viewport_index, i_vp));
        std::vector<int> fov_indices(dims.n_l1b_cols);
        std::iota(fov_indices.begin(),
                  fov_indices.end(),
                  std::distance(viewport_index.begin(),
                                std::ranges::find(viewport_index, i_vp)));
        std::vector<double> l1b_lat_vp(dims.n_l1b_rows * dims.n_l1b_cols);
        std::vector<double> l1b_lon_vp(dims.n_l1b_rows * dims.n_l1b_cols);
        for (int i {}; i < dims.n_l1b_rows; ++i) {
            for (int j {}; j < dims.n_l1b_cols; ++j) {
                const int idx { i * spatial_samples_per_image
                                + fov_indices[j] };
                const int idx_vp { i * dims.n_l1b_cols + j };
                l1b_lat_vp[idx_vp] = l1b_latitudes[idx];
                l1b_lon_vp[idx_vp] = l1b_longitudes[idx];
            }
        }

        // Generate the fraction L1C row and column indices. This
        // needs to be done only once per viewport.
        std::vector<double> row_indices(dims.n_l1b_rows * dims.n_l1b_cols);
        std::vector<double> col_indices(dims.n_l1b_rows * dims.n_l1b_cols);
        genL1CCoordinates(settings,
                          l1b_lat_vp,
                          l1b_lon_vp,
                          l1c_latitudes,
                          l1c_longitudes,
                          normal_vectors_r,
                          normal_vectors_c,
                          dims);

        // Do the interpolation
        std::vector<double> l1b_buf(dims.n_l1b_rows * dims.n_l1b_cols);
        std::vector<double> l1c_buf(dims.n_l1c_rows * dims.n_l1c_cols);
        std::vector<int> mpi_ranges {};
        genWorkRanges(intensity_bands_per_view, mpi.n_procs, mpi_ranges);
        for (int i_wave { mpi_ranges.at(mpi.rank) };
             i_wave < mpi_ranges.at(mpi.rank + 1);
             ++i_wave) {
            const auto loadToBuf { [&](const SharedArray<float>& l1b_in) {
                for (int i {}; i < dims.n_l1b_rows; ++i) {
                    for (int j {}; j < dims.n_l1b_cols; ++j) {
                        l1b_buf[i * dims.n_l1b_cols + j] =
                          l1b_in[(i * spatial_samples_per_image
                                  + fov_indices[j])
                                   * intensity_bands_per_view
                                 + i_wave];
                    }
                }
            } };
            const auto loadFromBuf { [&](SharedArray<float>& l1c_out) {
                for (int i {}; i < dims.n_l1c_rows; ++i) {
                    for (int j {}; j < dims.n_l1c_cols; ++j) {
                        l1c_out[((i * dims.n_l1c_cols + j) * number_of_views
                                 + i_vp)
                                  * intensity_bands_per_view
                                + i_wave] = l1c_buf[i * dims.n_l1c_cols + j];
                    }
                }
            } };
            loadToBuf(l1b_I);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_I);
            loadToBuf(l1b_I_noise);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_I_noise);
        }
        genWorkRanges(polarization_bands_per_view, mpi.n_procs, mpi_ranges);
        for (int i_wave { mpi_ranges.at(mpi.rank) };
             i_wave < mpi_ranges.at(mpi.rank + 1);
             ++i_wave) {
            auto loadToBuf { [&](const SharedArray<float>& l1b_in) {
                for (int i {}; i < dims.n_l1b_rows; ++i) {
                    for (int j {}; j < dims.n_l1b_cols; ++j) {
                        l1b_buf[i * dims.n_l1b_cols + j] =
                          l1b_in[(i * spatial_samples_per_image
                                  + fov_indices[j])
                                   * polarization_bands_per_view
                                 + i_wave];
                    }
                }
            } };
            auto loadFromBuf { [&](SharedArray<float>& l1c_out) {
                for (int i {}; i < dims.n_l1c_rows; ++i) {
                    for (int j {}; j < dims.n_l1c_cols; ++j) {
                        l1c_out[((i * dims.n_l1c_cols + j) * number_of_views
                                 + i_vp)
                                  * polarization_bands_per_view
                                + i_wave] = l1c_buf[i * dims.n_l1c_cols + j];
                    }
                }
            } };
            loadToBuf(l1b_I_polsample);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_I_polsample);
            loadToBuf(l1b_I_polsample_noise);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_I_polsample_noise);
            loadToBuf(l1b_q);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_q);
            loadToBuf(l1b_q_noise);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_q_noise);
            loadToBuf(l1b_u);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_u);
            loadToBuf(l1b_u_noise);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_u_noise);
            loadToBuf(l1b_dolp);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_dolp);
            loadToBuf(l1b_dolp_noise);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_dolp_noise);
            loadToBuf(l1b_aolp);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_aolp);
            loadToBuf(l1b_aolp_noise);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_aolp_noise);
        }
        if (mpi.rank == 0) {
            auto loadToBuf { [&](const SharedArray<float>& l1b_in) {
                for (int i {}; i < dims.n_l1b_rows; ++i) {
                    for (int j {}; j < dims.n_l1b_cols; ++j) {
                        l1b_buf[i * dims.n_l1b_cols + j] =
                          l1b_in[i * spatial_samples_per_image
                                 + fov_indices[j]];
                    }
                }
            } };
            auto loadFromBuf { [&](SharedArray<float>& l1c_out) {
                for (int i {}; i < dims.n_l1c_rows; ++i) {
                    for (int j {}; j < dims.n_l1c_cols; ++j) {
                        l1c_out[(i * dims.n_l1c_cols + j) * number_of_views
                                + i_vp] = l1c_buf[i * dims.n_l1c_cols + j];
                    }
                }
            } };
            loadToBuf(l1b_solar_zenith);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_sza);
            loadToBuf(l1b_solar_azimuth);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_saa);
            loadToBuf(l1b_sensor_zenith);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_vza);
            loadToBuf(l1b_sensor_azimuth);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            loadFromBuf(l1c_vaa);
            // Altitude does not have angular dependence. Add up
            // altitudes from all viewing angles and in the end divide
            // by the number of viewing angles.
            loadToBuf(l1b_altitude);
            interpolateAndBin(dims, l1b_buf, l1c_buf);
            if (i_vp == 0) {
                std::ranges::fill(l1c_altitude, 0.0f);
            }
            for (int i {}; i < dims.n_l1c_rows * dims.n_l1c_cols; ++i) {
                l1c_altitude[i] += l1c_buf[i];
            }
        }
    }
#ifdef USE_MPI
    MPI_Barrier(mpi.comm);
#endif

    // Write to NetCDF file
    if (mpi.rank != 0) {
        return;
    }
    NcObject nc_l1c { netCDF::NcFile::replace, settings.l1c.l1c_out };
    nc_l1c.putAtt("processing_level", "L1C");
    nc_l1c.putAtt("title", "PACE SPEXone Level-1C data");
    nc_l1c.putAtt("bin_size_at_nadir", "2.6 km");

    // Dimensions
    const auto nc_bins_along_track { nc_l1c.addDim("bins_along_track",
                                                   dims.n_l1c_rows) };
    const auto nc_bins_across_track { nc_l1c.addDim("bins_across_track",
                                                    dims.n_l1c_cols) };
    const auto nc_number_of_views { nc_l1c.addDim("number_of_views",
                                                  number_of_views) };
    const auto nc_intensity_bands_per_view { nc_l1c.addDim(
      "intensity_bands_per_view", intensity_bands_per_view) };
    const auto nc_polarization_bands_per_view { nc_l1c.addDim(
      "polarization_bands_per_view", polarization_bands_per_view) };
    const std::vector<netCDF::NcDim> nc_int_dims {
        nc_bins_along_track,
        nc_bins_across_track,
        nc_number_of_views,
        nc_intensity_bands_per_view
    };
    const std::vector<netCDF::NcDim> nc_pol_dims {
        nc_bins_along_track,
        nc_bins_across_track,
        nc_number_of_views,
        nc_polarization_bands_per_view
    };

    // Sensor view bands
    auto nc_sensor_views_bands { nc_l1c.addGroup("sensor_views_bands") };
    auto nc_view_angles { nc_sensor_views_bands.addVar(
      "view_angles", netCDF::ncFloat, nc_number_of_views) };
    nc_view_angles.putAtt("long_name", "along-track view angles for sensor");
    nc_view_angles.putAtt("units", "degrees");
    nc_view_angles.putAtt("_FillValue", netCDF::ncFloat, -999.f);
    nc_view_angles.putAtt("valid_min", netCDF::ncFloat, -89.0);
    nc_view_angles.putAtt("valid_max", netCDF::ncFloat, 89.0);
    std::vector<float> vp_angles(number_of_views, 0);
    std::transform(viewport_idx_to_angle.cbegin(),
                   viewport_idx_to_angle.cend(),
                   vp_angles.begin(),
                   [](const auto& pair) { return pair.second; });
    nc_view_angles.putVar(vp_angles.data());
    auto nc_intensity_wavelengths { nc_sensor_views_bands.addVar(
      "intensity_wavelengths",
      netCDF::ncFloat,
      { nc_number_of_views, nc_intensity_bands_per_view }) };
    nc_intensity_wavelengths.putAtt(
      "long_name", "intensity field center wavelengths at each view");
    nc_intensity_wavelengths.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_intensity_wavelengths.putAtt(
      "valid_min", netCDF::ncFloat, L1B::wavelengths_valid_min);
    nc_intensity_wavelengths.putAtt(
      "valid_max", netCDF::ncFloat, L1B::wavelengths_valid_max);
    nc_intensity_wavelengths.putAtt("units", "nm");
    nc_intensity_wavelengths.putVar(intensity_wavelengths.data());
    auto nc_intensity_bandpasses { nc_sensor_views_bands.addVar(
      "intensity_bandpasses",
      netCDF::ncFloat,
      { nc_number_of_views, nc_intensity_bands_per_view }) };
    nc_intensity_bandpasses.putAtt("long_name",
                                   "intensity field bandpasses at each view");
    nc_intensity_bandpasses.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_intensity_bandpasses.putAtt("valid_min", netCDF::ncFloat, 2.5f);
    nc_intensity_bandpasses.putAtt("valid_max", netCDF::ncFloat, 100.0f);
    nc_intensity_bandpasses.putAtt("units", "nm");
    auto nc_polarization_wavelengths { nc_sensor_views_bands.addVar(
      "polarization_wavelengths",
      netCDF::ncFloat,
      { nc_number_of_views, nc_polarization_bands_per_view }) };
    nc_polarization_wavelengths.putAtt(
      "long_name", "wavelength at center of polarization bands");
    nc_polarization_wavelengths.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_polarization_wavelengths.putAtt(
      "valid_min", netCDF::ncFloat, L1B::wavelengths_valid_min);
    nc_polarization_wavelengths.putAtt(
      "valid_max", netCDF::ncFloat, L1B::wavelengths_valid_max);
    nc_polarization_wavelengths.putAtt("units", "nm");
    nc_polarization_wavelengths.putVar(polarization_wavelengths.data());
    auto nc_polarization_bandpasses { nc_sensor_views_bands.addVar(
      "polarization_bandpasses",
      netCDF::ncFloat,
      { nc_number_of_views, nc_polarization_bands_per_view }) };
    nc_polarization_bandpasses.putAtt(
      "long_name", "polarization field bandpasses at each view");
    nc_polarization_bandpasses.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_polarization_bandpasses.putAtt(
      "valid_min", netCDF::ncFloat, L1B::wavelengths_valid_min);
    nc_polarization_bandpasses.putAtt(
      "valid_max", netCDF::ncFloat, L1B::wavelengths_valid_max);
    nc_polarization_bandpasses.putAtt("units", "nm");
    auto nc_intensity_F0 { nc_sensor_views_bands.addVar(
      "intensity_F0",
      netCDF::ncFloat,
      { nc_number_of_views, nc_intensity_bands_per_view }) };
    nc_intensity_F0.putAtt("long_name",
                           "intensity field bandpasses at each view");
    nc_intensity_F0.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_intensity_F0.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    nc_intensity_F0.putAtt("valid_max", netCDF::ncFloat, 900.0f);
    nc_intensity_F0.putAtt("units", "W.m-2.um-1");
    auto nc_polarization_F0 { nc_sensor_views_bands.addVar(
      "polarization_F0",
      netCDF::ncFloat,
      { nc_number_of_views, nc_polarization_bands_per_view }) };
    nc_polarization_bandpasses.putAtt(
      "long_name", "polarization field bandpasses at each view");
    nc_polarization_bandpasses.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_polarization_bandpasses.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    nc_polarization_bandpasses.putAtt("valid_max", netCDF::ncFloat, 900.0f);
    nc_polarization_bandpasses.putAtt("units", "W.m-2.um-1");

    // Bin attributes
    auto nc_bin_attributes { nc_l1c.addGroup("bin_attributes") };
    auto nc_nadir_view_time { nc_bin_attributes.addVar(
      "nadir_view_time", netCDF::ncDouble, nc_bins_along_track) };
    nc_nadir_view_time.putAtt("long_name", "time bin was viewed at nadir view");
    nc_nadir_view_time.putAtt("_FillValue", netCDF::ncDouble, -999.0);
    nc_nadir_view_time.putAtt("valid_min", netCDF::ncDouble, 0.0);
    nc_nadir_view_time.putAtt("valid_max", netCDF::ncDouble, 86400.999);
    nc_nadir_view_time.putAtt("units", "seconds");
    nc_nadir_view_time.putVar(image_time.data());

    auto nc_view_time_offsets { nc_bin_attributes.addVar(
      "view_time_offsets", netCDF::ncDouble, nc_bins_along_track) };
    nc_view_time_offsets.putAtt("long_name",
                                "time offsets of views from nadir view");
    nc_view_time_offsets.putAtt("_FillValue", netCDF::ncDouble, -999.0);
    nc_view_time_offsets.putAtt("valid_min", netCDF::ncDouble, -200.0);
    nc_view_time_offsets.putAtt("valid_max", netCDF::ncDouble, 200.0);
    nc_view_time_offsets.putAtt("units", "seconds");

    // Geolocation data
    auto nc_geolocation_data { nc_l1c.addGroup("geolocation_data") };
    // For converting float arrays to integer arrays and scaling
    std::vector<int16_t> buf_16 {};
    const std::vector<netCDF::NcDim> nc_latlon_dims {
        nc_bins_along_track,
        nc_bins_across_track,
    };
    // Like lat/lon but with angular dependence
    const std::vector<netCDF::NcDim> nc_ang_dims { nc_bins_along_track,
                                                   nc_bins_across_track,
                                                   nc_number_of_views };
    // To avoid loss of accuracy, reread the latitudes and longitudes
    // from the original grid file.
    NcObject nc_l1c_grid { netCDF::NcFile::read, settings.l1c.l1c_in };
    nc_l1c_grid.getGroup("geolocation_data")
      .getVar("latitude")
      .getVar(l1c_latitudes.data());
    nc_l1c_grid.getGroup("geolocation_data")
      .getVar("longitude")
      .getVar(l1c_longitudes.data());
    auto nc_latitude { nc_geolocation_data.addVar(
      "latitude", netCDF::ncFloat, nc_latlon_dims) };
    nc_latitude.putAtt("_FillValue", netCDF::ncFloat, -999.9f);
    nc_latitude.putAtt("long_name", "latitudes of bin locations");
    nc_latitude.putAtt("valid_min", netCDF::ncFloat, -90.0f);
    nc_latitude.putAtt("valid_max", netCDF::ncFloat, 90.0f);
    nc_latitude.putAtt("units", "degrees_north");
    nc_latitude.putVar(l1c_latitudes.data());
    auto nc_longitude { nc_geolocation_data.addVar(
      "longitude", netCDF::ncFloat, nc_latlon_dims) };
    nc_longitude.putAtt("_FillValue", netCDF::ncFloat, -999.9f);
    nc_longitude.putAtt("long_name", "longitudes of bin locations");
    nc_longitude.putAtt("valid_min", netCDF::ncFloat, -180.0f);
    nc_longitude.putAtt("valid_max", netCDF::ncFloat, 180.0f);
    nc_longitude.putAtt("units", "degrees_east");
    nc_longitude.putVar(l1c_longitudes.data());
    auto nc_altitude { nc_geolocation_data.addVar(
      "altitude", netCDF::ncShort, nc_latlon_dims) };
    nc_altitude.putAtt("_FillValue", netCDF::ncShort, -32768);
    nc_altitude.putAtt("long_name", "altitude at bin locations");
    nc_altitude.putAtt("valid_min", netCDF::ncShort, -1000);
    nc_altitude.putAtt("valid_max", netCDF::ncShort, 10000);
    nc_altitude.putAtt("units", "m");
    nc_altitude.putAtt("scale_factor", netCDF::ncFloat, 1.0f);
    nc_altitude.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    buf_16.resize(l1c_altitude.size());
    for (int i {}; i < static_cast<int>(buf_16.size()); ++i) {
        buf_16[i] = static_cast<int16_t>(l1c_altitude[i] / number_of_views);
    }
    nc_altitude.putVar(buf_16.data());
    auto nc_altitude_var { nc_geolocation_data.addVar(
      "altitude_variability", netCDF::ncShort, nc_latlon_dims) };
    nc_altitude_var.putAtt("_FillValue", netCDF::ncShort, -32768);
    nc_altitude_var.putAtt("long_name",
                           "RMS variability of altitude at bin locations");
    nc_altitude_var.putAtt("valid_min", netCDF::ncShort, 0);
    nc_altitude_var.putAtt("valid_max", netCDF::ncShort, 1000.0);
    nc_altitude_var.putAtt("units", "meters");
    nc_altitude_var.putAtt("scale_factor", netCDF::ncFloat, 1.0f);
    nc_altitude_var.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    auto nc_sza { nc_geolocation_data.addVar(
      "solar_zenith", netCDF::ncShort, nc_ang_dims) };
    nc_sza.putAtt("_FillValue", netCDF::ncShort, -32768);
    nc_sza.putAtt("long_name", "Solar zenith angle at bin locations");
    nc_sza.putAtt("valid_min", netCDF::ncShort, 0);
    nc_sza.putAtt("valid_max", netCDF::ncShort, 18000);
    nc_sza.putAtt("units", "degrees");
    nc_sza.putAtt("scale_factor", netCDF::ncFloat, 1.0f / geo_scale_factor);
    nc_sza.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    floatToShort(l1c_sza, buf_16);
    nc_sza.putVar(buf_16.data());
    auto nc_saa { nc_geolocation_data.addVar(
      "solar_azimuth", netCDF::ncShort, nc_ang_dims) };
    nc_saa.putAtt("_FillValue", netCDF::ncShort, -32768);
    nc_saa.putAtt("long_name", "Solar azimuth angle at bin locations");
    nc_saa.putAtt("valid_min", netCDF::ncShort, -18000);
    nc_saa.putAtt("valid_max", netCDF::ncShort, 18000);
    nc_saa.putAtt("units", "degrees");
    nc_saa.putAtt("scale_factor", netCDF::ncFloat, 1.0f / geo_scale_factor);
    nc_saa.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    floatToShort(l1c_saa, buf_16);
    nc_saa.putVar(buf_16.data());
    auto nc_vza { nc_geolocation_data.addVar(
      "sensor_zenith", netCDF::ncShort, nc_ang_dims) };
    nc_vza.putAtt("_FillValue", netCDF::ncShort, -32768);
    nc_vza.putAtt("long_name", "Sensor zenith angle at bin locations");
    nc_vza.putAtt("valid_min", netCDF::ncShort, 0);
    nc_vza.putAtt("valid_max", netCDF::ncShort, 18000);
    nc_vza.putAtt("units", "degrees");
    nc_vza.putAtt("scale_factor", netCDF::ncFloat, 1.0f / geo_scale_factor);
    nc_vza.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    floatToShort(l1c_vza, buf_16);
    nc_vza.putVar(buf_16.data());
    auto nc_vaa { nc_geolocation_data.addVar(
      "sensor_azimuth", netCDF::ncShort, nc_ang_dims) };
    nc_vaa.putAtt("_FillValue", netCDF::ncShort, -32768);
    nc_vaa.putAtt("long_name", "Sensor azimuth angle at bin locations");
    nc_vaa.putAtt("valid_min", netCDF::ncShort, -18000);
    nc_vaa.putAtt("valid_max", netCDF::ncShort, 18000);
    nc_vaa.putAtt("units", "degrees");
    nc_vaa.putAtt("scale_factor", netCDF::ncFloat, 1.0f / geo_scale_factor);
    nc_vaa.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    floatToShort(l1c_vaa, buf_16);
    nc_vaa.putVar(buf_16.data());
    // Finally, the scattering angle is computed from the solar and
    // sensor viewing angles.
    auto nc_scattering_angle { nc_geolocation_data.addVar(
      "scattering_angle", netCDF::ncShort, nc_ang_dims) };
    nc_scattering_angle.putAtt("_FillValue", netCDF::ncShort, -32768);
    nc_scattering_angle.putAtt("long_name",
                               "scattering angle at bin locations");
    nc_scattering_angle.putAtt("valid_min", netCDF::ncShort, 0);
    nc_scattering_angle.putAtt("valid_max", netCDF::ncShort, 18000);
    nc_scattering_angle.putAtt("units", "degrees");
    nc_scattering_angle.putAtt(
      "scale_factor", netCDF::ncFloat, 1.0f / geo_scale_factor);
    nc_scattering_angle.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    SharedArray<float> scattering_angle { dims.n_l1c_rows * dims.n_l1c_cols
                                          * number_of_views };
    for (int i {}; i < static_cast<int>(scattering_angle.size()); ++i) {
        const double theta { math::deg_to_rad * l1c_vza[i] };
        const double theta_s { math::deg_to_rad * l1c_sza[i] };
        const double phi { math::deg_to_rad * l1c_vaa[i] };
        const double phi_s { math::deg_to_rad * l1c_saa[i] };
        const double u { std::cos(theta + std::numbers::pi) };
        const double u_s { std::cos(theta_s) };
        const double cos_alpha {
            u * u_s
            + std::sqrt(1 - u * u) * std::sqrt(1 - u_s * u_s)
                * std::cos(phi + std::numbers::pi - phi_s)
        };
        scattering_angle[i] =
          static_cast<float>(std::acos(cos_alpha) / math::deg_to_rad);
    }
    floatToShort(scattering_angle, buf_16);
    nc_scattering_angle.putVar(buf_16.data());

    // Observation data
    auto nc_observation_data { nc_l1c.addGroup("observation_data") };
    auto nc_obs_per_view { nc_geolocation_data.addVar(
      "obs_per_view", netCDF::ncShort, nc_ang_dims) };
    nc_obs_per_view.putAtt("long_name",
                           "observations contributing to bin from each view");
    nc_obs_per_view.putAtt("valid_min", netCDF::ncShort, 0);
    nc_obs_per_view.putAtt("valid_max", netCDF::ncShort, 999);
    nc_obs_per_view.putAtt("units", "1");
    auto nc_QC_bitwise { nc_observation_data.addVar(
      "QC_bitwise", netCDF::ncFloat, nc_int_dims) };
    nc_QC_bitwise.putAtt("long_name", "bitwise quality indicator");
    nc_QC_bitwise.putAtt("_FillValue", netCDF::ncFloat, 0.0f);
    nc_QC_bitwise.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    nc_QC_bitwise.putAtt("valid_max", netCDF::ncFloat, 65536.0f);
    nc_QC_bitwise.putAtt("units", "1");
    auto nc_QC { nc_observation_data.addVar(
      "QC", netCDF::ncFloat, nc_int_dims) };
    nc_QC.putAtt("long_name", "quality indicator");
    nc_QC.putAtt("_FillValue", netCDF::ncFloat, 0.0f);
    nc_QC.putAtt("valid_min", netCDF::ncShort, 0.0f);
    nc_QC.putAtt("valid_max", netCDF::ncShort, 10.0f);
    nc_QC.putAtt("units", "1");
    auto nc_I { nc_observation_data.addVar("I", netCDF::ncFloat, nc_int_dims) };
    nc_I.putAtt("long_name", "I stokes vector component");
    nc_I.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
    nc_I.putAtt("units", "W.m-2.sr-1.um-1");
    nc_I.putAtt("ancillary_variables", "I_noise");
    nc_I.putVar(l1c_I.data());
    auto nc_I_noise { nc_observation_data.addVar(
      "I_noise", netCDF::ncFloat, nc_int_dims) };
    nc_I_noise.putAtt("long_name", "random noise of I in bin");
    nc_I_noise.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
    nc_I_noise.putAtt("units", "W.m-2.sr-1.um-1");
    nc_I_noise.putVar(l1c_I_noise.data());
    auto nc_QC_polsample_bitwise { nc_observation_data.addVar(
      "QC_polsample_bitwise", netCDF::ncFloat, nc_pol_dims) };
    nc_QC_polsample_bitwise.putAtt("long_name", "bitwise quality indicator");
    nc_QC_polsample_bitwise.putAtt("_FillValue", netCDF::ncFloat, 0.0f);
    nc_QC_polsample_bitwise.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    nc_QC_polsample_bitwise.putAtt("valid_max", netCDF::ncFloat, 65536.0f);
    nc_QC_polsample_bitwise.putAtt("units", "1");
    auto nc_QC_polsample { nc_observation_data.addVar(
      "QC_polsample", netCDF::ncFloat, nc_pol_dims) };
    nc_QC_polsample.putAtt("long_name", "quality indicator");
    nc_QC_polsample.putAtt("_FillValue", netCDF::ncFloat, 0.0f);
    nc_QC_polsample.putAtt("valid_min", netCDF::ncShort, 0.0f);
    nc_QC_polsample.putAtt("valid_max", netCDF::ncShort, 10.0f);
    nc_QC_polsample.putAtt("units", "1");
    auto nc_I_polsample { nc_observation_data.addVar(
      "I_polsample", netCDF::ncFloat, nc_pol_dims) };
    nc_I_polsample.putAtt(
      "long_name",
      "I Stokes vector component at polarimeter spectral sampling");
    nc_I_polsample.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_I_polsample.putAtt("units", "W.m-2.sr-1.um-1");
    nc_I_polsample.putAtt("ancillary_variables", "I_pol_noise");
    nc_I_polsample.putVar(l1c_I_polsample.data());
    auto nc_I_polsample_noise { nc_observation_data.addVar(
      "I_polsample_noise", netCDF::ncFloat, nc_pol_dims) };
    nc_I_polsample_noise.putAtt("long_name",
                                "random noise of I_polsample in bin");
    nc_I_polsample_noise.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_I_polsample_noise.putAtt("units", "W.m-2.sr-1.um-1");
    nc_I_polsample_noise.putVar(l1c_I_polsample_noise.data());
    auto nc_q { nc_observation_data.addVar(
      "Q_over_I", netCDF::ncFloat, nc_pol_dims) };
    nc_q.putAtt("long_name", "Q over I (little q) Stokes vector component");
    nc_q.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_q.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    nc_q.putAtt("valid_max", netCDF::ncFloat, 1.0f);
    nc_q.putAtt("units", "1");
    nc_q.putAtt("ancillary_variables", "q_noise");
    nc_q.putVar(l1c_q.data());
    auto nc_q_noise { nc_observation_data.addVar(
      "Q_over_I_noise", netCDF::ncFloat, nc_pol_dims) };
    nc_q_noise.putAtt("long_name", "random noise of Q over I in bin");
    nc_q_noise.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_q_noise.putAtt("units", "1");
    nc_q_noise.putVar(l1c_q_noise.data());
    auto nc_u { nc_observation_data.addVar(
      "U_over_I", netCDF::ncFloat, nc_pol_dims) };
    nc_u.putAtt("long_name", "U over I (little u) Stokes vector component");
    nc_u.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_u.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    nc_u.putAtt("valid_max", netCDF::ncFloat, 1.0f);
    nc_u.putAtt("units", "1");
    nc_u.putAtt("ancillary_variables", "u_noise");
    nc_u.putVar(l1c_u.data());
    auto nc_u_noise { nc_observation_data.addVar(
      "U_over_I_noise", netCDF::ncFloat, nc_pol_dims) };
    nc_u_noise.putAtt("long_name", "random noise of U over I in bin");
    nc_u_noise.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_u_noise.putAtt("units", "1");
    nc_u_noise.putVar(l1c_u_noise.data());
    auto nc_dolp { nc_observation_data.addVar(
      "DOLP", netCDF::ncFloat, nc_pol_dims) };
    nc_dolp.putAtt("long_name", "degree of linear polarization");
    nc_dolp.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_dolp.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    nc_dolp.putAtt("valid_max", netCDF::ncFloat, 1.0f);
    nc_dolp.putAtt("units", "1");
    nc_dolp.putAtt("ancillary_variables", "DoLP_noise");
    nc_dolp.putVar(l1c_dolp.data());
    auto nc_dolp_noise { nc_observation_data.addVar(
      "DOLP_noise", netCDF::ncFloat, nc_pol_dims) };
    nc_dolp_noise.putAtt("long_name", "random noise of DoLP in bin");
    nc_dolp_noise.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_dolp_noise.putAtt("units", "1");
    nc_dolp_noise.putVar(l1c_dolp_noise.data());
    auto nc_aolp { nc_observation_data.addVar(
      "AOLP", netCDF::ncFloat, nc_pol_dims) };
    nc_aolp.putAtt("long_name", "angle of linear polarization");
    nc_aolp.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_aolp.putAtt("valid_min", netCDF::ncFloat, -90.0f);
    nc_aolp.putAtt("valid_max", netCDF::ncFloat, 90.0f);
    nc_aolp.putAtt("units", "degrees");
    nc_aolp.putAtt("ancillary_variables", "AoLP_noise");
    nc_aolp.putVar(l1c_aolp.data());
    auto nc_aolp_noise { nc_observation_data.addVar(
      "AOLP_noise", netCDF::ncFloat, nc_pol_dims) };
    nc_aolp_noise.putAtt("long_name", "random noise of AoLP in bin");
    nc_aolp_noise.putAtt("_FillValue", netCDF::ncFloat, -999.0f);
    nc_aolp_noise.putAtt("units", "degrees");
    nc_aolp_noise.putVar(l1c_aolp_noise.data());
}

} // namespace spex
