// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "nc_object.h"

#include "date.h"
#include "io.h"
#include "spex_settings.h"

#include <spdlog/spdlog.h>

namespace spex {

NcObject::NcObject(const netCDF::NcFile::FileMode fmode,
                   const std::string& filename)
  : fmode { fmode }
{
    if (!filename.empty()) {
        open(filename);
    }
}

auto NcObject::open(const std::string& filename) -> void
{
    close();
    this->filename = filename;
    netCDF::NcFile::open(filename, fmode);
    if (fmode != netCDF::NcFile::read) {
        putAtt("Conventions", "CF-1.6");
        putAtt("institution", "SRON Netherlands Institute for Space Research");
        putAtt("instrument", "SPEXone");
        putAtt("processor_version", SPEXONE_PROJECT_VERSION);
        putAtt("product_name", filename);
        putAtt("project", "PACE Project");
        putAtt("creator_email", "SPEXone-MPC@sron.nl");
        putAtt("creator_name", "SRON/Earth Science");
        putAtt("creator_url",
               "https://www.sron.nl/missions-earth/pace-spexone");
        putAtt("publisher_email", "SPEXone-MPC@sron.nl");
        putAtt("publisher_name", "SRON/Earth Science");
        putAtt("publisher_url",
               "https://www.sron.nl/missions-earth/pace-spexone");
        putAtt("date_created", getDateAndTime());
        putAtt("git_commit", SPEXONE_GIT_COMMIT_ABBREV);
    }
}

[[nodiscard]] auto NcObject::hasBeenOpened() const -> bool
{
    return !filename.empty();
}

auto NcObject::reOpen() -> void
{
    if (!hasBeenOpened()) {
        throw std::runtime_error {
            "cannot reopen a NetCDF file that has not been opened yet"
        };
    }
    netCDF::NcFile::open(filename, fmode);
}

auto NcObject::writeReducedConfig(const SPEXSettings& settings) -> void
{
    auto var { addVar("processor_configuration", netCDF::ncString) };
    var.putAtt("comment",
               "configuration parameters used during the processor run that "
               "produced this file");
    std::string conf_str { settings.getReducedConfig().c_str() };
    const char* conf_char { conf_str.c_str() };
    var.putVar(&conf_char);
}

auto NcObject::getGroup(const std::string& path) -> netCDF::NcGroup
{
    const std::vector<std::string> name_list { splitString(path, '/') };
    auto name_it { name_list.cbegin() };
    // The first group is taken from NcFile and subsequent groups, if
    // any, are taken as subgroups of the current group.
    netCDF::NcGroup group { this->NcFile::getGroup(*name_it) };
    while (++name_it != name_list.cend() && !group.isNull()) {
        group = group.getGroup(*name_it);
    }
    return group;
}

NcObject::~NcObject()
{
    if (!filename.empty()) {
        if (fmode != netCDF::NcFile::read) {
            putAtt("date_finalized", getDateAndTime());
        }
    }
}

} // namespace spex
