// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "nonlin_gen.h"

#include "bspline.h"
#include "functions.h"
#include "io.h"
#include "l1a_metadata.h"
#include "linalg.h"
#include "linear_spline.h"
#include "lininv.h"
#include "spex_math.h"

#include <gadfit/lm_solver.h>
#include <numeric>

namespace spex {

auto NonlinGen::processInit() -> void
{
    ckd.nonlin.model = settings.nonlin.model;
    genWorkRanges(ckd.npix, mpi.n_procs, pix_ranges);
    if (ckd.nonlin.model == NonlinModel::sigmoidal) {
        return;
    }
    // Create the fitting B-spline, meanwhile leaving all essenntial information
    // on the 'nonlin'-prefixed variables in the CKD structure.
    ckd.dim_nonlin_knot = settings.nonlin.knots.size();
    ckd.nonlin_knots.resize(ckd.dim_nonlin_knot);
    for (int i {}; i < static_cast<int>(ckd.nonlin_knots.size()); ++i) {
        ckd.nonlin_knots[i] = settings.nonlin.knots[i];
    }
    ckd.nonlin_order = settings.nonlin.order;
    b = std::make_unique<Bspline>(
      ckd.nonlin_order, ckd.dim_nonlin_knot, ckd.nonlin_knots.data());
    ckd.dim_nonlin_spline = b->nspline;
    abscissa.resize(nL1A());
    for (int i_l1a {}; i_l1a < nL1A(); ++i_l1a) {
        abscissa[i_l1a] =
          l1a_manager->l1a_products[i_l1a].exposure_time
          * l1a_manager->l1a_products[i_l1a].meta->illumination_level;
    }
    ckd.nonlin_signal_scale.resize(ckd.npix);
    nonlin_lin_slope.resize(ckd.npix);
    nonlin_lin_chi2.resize(ckd.npix);
    ckd.nonlin_fit.resize(ckd.npix * ckd.dim_nonlin_spline);
    nonlin_chi2.resize(ckd.npix);
}

auto NonlinGen::addSignalInduced(
  const bool active,
  std::vector<std::reference_wrapper<L1A>>& l1a_products) const -> void
{
    spdlog::info("Applying signal induced offset");
    std::vector<int> mpi_ranges {};
    genWorkRanges(
      static_cast<int>(l1a_products.size()), mpi.n_procs, mpi_ranges);
    for (int i_l1a { mpi_ranges.at(mpi.rank) };
         i_l1a < mpi_ranges.at(mpi.rank + 1);
         ++i_l1a) {
        for (int i_row {}; i_row < ckd.dim_detector_spat; ++i_row) {
            // Average across columns for this row
            double average {};
            // Number of values to be averaged can be less than the
            // detector dimension if there are bad pixels.
            int n_values {};
            for (int i_col {}; i_col < ckd.dim_detector_spec; ++i_col) {
                const int idx { i_row * ckd.dim_detector_spec + i_col };
                if (!ckd.pixel_mask[idx]) {
                    average +=
                      batches.front().l1a_products[i_l1a].get().image[idx];
                    ++n_values;
                }
            }
            constexpr double rel_factor { 0.17 };
            const double fac { active
                                 ? rel_factor * std::sqrt(average / n_values)
                                 : 0.0 };
            for (int i_col {}; i_col < ckd.dim_detector_spec; ++i_col) {
                const int idx { i_row * ckd.dim_detector_spec + i_col };
                batches.front().l1a_products[i_l1a].get().image[idx] +=
                  ckd.dark.current[idx]
                    * batches.front().l1a_products[i_l1a].get().exposure_time
                  - fac;
            }
        }
    }
}

// Sort measurements into batches such that within each batch the
// reference diode value is more or less constant. Preliminary to step
// 4.
static auto groupByReferenceSignal(
  const std::vector<std::reference_wrapper<L1A>>& l1a_products)
  -> std::vector<std::vector<std::reference_wrapper<L1A>>>
{
    std::vector<std::reference_wrapper<L1A>> l1a_refs { l1a_products.begin(),
                                                        l1a_products.end() };
    std::ranges::sort(l1a_refs, [](const auto& l1a_1, const auto& l1a_2) {
        return l1a_1.get().meta->reference_signal
               < l1a_2.get().meta->reference_signal;
    });
    // Group by light intensity
    std::vector<std::vector<std::reference_wrapper<L1A>>> l1a_grouped_by_int {};
    auto l1a_first { l1a_refs.begin() };
    auto l1a_last { l1a_refs.begin() };
    while (l1a_last++ != l1a_refs.end()) {
        const auto& meta_first { l1a_first->get().meta };
        const auto& meta_last { l1a_last != l1a_refs.end()
                                  ? l1a_last->get().meta
                                  : meta_first };
        const auto rel_ref_signal_diff { std::abs(
          (meta_first->reference_signal - meta_last->reference_signal)
          / meta_first->reference_signal) };
        constexpr double rel_ref_signal_diff_thres { 0.01 };
        if (l1a_last == l1a_refs.end()
            || meta_first->filename.find("_lin_") == std::string::npos
            || meta_last->filename.find("_lin_") == std::string::npos
            || rel_ref_signal_diff > rel_ref_signal_diff_thres) {
            // Do not create a group if there are 3 or less
            // measurements (exposure times)
            if (std::ranges::distance(l1a_first, l1a_last) > 3) {
                l1a_grouped_by_int.push_back({ l1a_first, l1a_last });
            }
            l1a_first = l1a_last;
        }
    }
    for (auto& group : l1a_grouped_by_int) {
        std::ranges::sort(group, [](const auto& l1a_1, const auto& l1a_2) {
            return l1a_1.get().exposure_time < l1a_2.get().exposure_time;
        });
    }
    // Now all measurements are grouped by light level and exposure time
    return l1a_grouped_by_int;
}

auto NonlinGen::computeSignalRates(
  const std::vector<std::reference_wrapper<L1A>>& l1a_products) const -> void
{
    spdlog::info("Computing signal rates");
    std::vector<std::reference_wrapper<L1A>> l1a_refs { l1a_products.begin(),
                                                        l1a_products.end() };
    // L1A products grouped by light level (reference diode value)
    const auto l1a_grouped_by_int { groupByReferenceSignal(l1a_products) };
    resize(ckd.nonlin.signal_rates,
           mpi,
           static_cast<int>(l1a_grouped_by_int.size()),
           ckd.npix);
    resize(ckd.nonlin.reference_rates,
           mpi,
           static_cast<int>(l1a_grouped_by_int.size()),
           ckd.npix);
    for (int i_group {}; i_group < static_cast<int>(l1a_grouped_by_int.size());
         ++i_group) {
        const int n_l1a { static_cast<int>(
          l1a_grouped_by_int.at(i_group).size()) };
        const int n_l1a_min { static_cast<int>(0.7 * n_l1a) };
        std::vector<double> exposure_times(n_l1a);
        // Compute signal rate per pixel with relative least squares
        // regression
        for (int i_pix { pix_ranges.at(mpi.rank) };
             i_pix < pix_ranges.at(mpi.rank + 1);
             ++i_pix) {
            double Q_00 {};
            double Q_11 {};
            double Q_22 {};
            for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
                exposure_times[i_l1a] =
                  l1a_grouped_by_int[i_group][i_l1a].get().exposure_time;
            }
            double sigma_min { 1e100 };
            int data_point_count {};
            for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
                double w;
                if (i_l1a == 0) {
                    w = 2 * (exposure_times[1] - exposure_times.front());
                } else if (i_l1a == n_l1a - 1) {
                    w = 2 * (exposure_times.back() - exposure_times[n_l1a - 2]);
                } else {
                    w = exposure_times[i_l1a + 1] - exposure_times[i_l1a - 1];
                }
                const double w_over_exp { w / exposure_times[i_l1a] };
                const double w_over_exp2 { w_over_exp / exposure_times[i_l1a] };
                Q_00 += w;
                const double& counts {
                    l1a_grouped_by_int[i_group][i_l1a].get().image[i_pix]
                };
                Q_11 += w_over_exp * counts;
                Q_22 += w_over_exp2 * counts * counts;
                const double signal_rate { Q_11 / Q_00 };
                const int degrees_of_freedom { i_l1a - 1 };
                const double chi2 { Q_22 - Q_00 * signal_rate * signal_rate };
                const double sigma { chi2 / degrees_of_freedom / Q_00 };
                if (i_l1a > n_l1a_min && sigma < sigma_min) {
                    sigma_min = sigma;
                    ckd.nonlin.signal_rates[i_group][i_pix] = signal_rate;
                    data_point_count = i_l1a + 1;
                }
            }
            for (int i_l1a {}; i_l1a < data_point_count; ++i_l1a) {
                ckd.nonlin.reference_rates[i_group][i_pix] +=
                  l1a_grouped_by_int[i_group][i_l1a]
                    .get()
                    .meta->reference_signal;
            }
            ckd.nonlin.reference_rates[i_group][i_pix] /= data_point_count;
        }
    }
}

auto NonlinGen::computeRefCurrent() const -> void
{
    spdlog::info("Computing reference diode conversion factors f1 and f2");
    constexpr int n_parameters { 2 };
    // Eq. (81)
    ckd.nonlin.f1 = { true, mpi, ckd.npix };
    ckd.nonlin.f2 = { true, mpi, ckd.npix };
    std::vector<double> I_c(ckd.nonlin.signal_rates.size());
    const std::vector<double> weights(ckd.nonlin.signal_rates.size(), 1.0);
    std::vector<double> jacobian(ckd.nonlin.signal_rates.size() * n_parameters);
    for (int i_pix { pix_ranges.at(mpi.rank) };
         i_pix < pix_ranges.at(mpi.rank + 1);
         ++i_pix) {
        // Loop over light intensities
        for (int i_int {};
             i_int < static_cast<int>(ckd.nonlin.signal_rates.size());
             ++i_int) {
            I_c[i_int] =
              ckd.nonlin.signal_rates[i_int][i_pix] - ckd.dark.current[i_pix];
            const double& ref_rate { ckd.nonlin.reference_rates[i_int][i_pix] };
            jacobian[i_int * n_parameters + 0] = ref_rate;
            jacobian[i_int * n_parameters + 1] = ref_rate * std::sqrt(ref_rate);
        }
        std::vector<double> state_vector(n_parameters, 0.0);
        dppsv(I_c, weights, jacobian, state_vector);
        ckd.nonlin.f1[i_pix] = state_vector[0];
        ckd.nonlin.f2[i_pix] =
          state_vector[1]
          / (ckd.nonlin.f1[i_pix] * std::sqrt(ckd.nonlin.f1[i_pix]));
    }
}

auto NonlinGen::determineMaximumSignals(
  const std::vector<std::reference_wrapper<L1A>>& l1a_products) const -> void
{
    spdlog::info("Computing saturated signal parameters (m0, m1, m2)");
    std::vector<double> values_for_median(ckd.npix);
    for (int i {}; i < ckd.npix; ++i) {
        values_for_median[i] = ckd.dark.offset_long[i];
    }
    const double dark_offset_median { median(values_for_median) };
    const double clipping_threshold {
        1000.0 - l1a_products.front().get().digital_offset_single
        - dark_offset_median
    };
    // Photodiode reference values
    std::vector<double> I_ref {};
    for (int i_l1a {}; i_l1a < static_cast<int>(l1a_products.size()); ++i_l1a) {
        // Only consider long exposure times where data is saturated
        constexpr double exposure_thres { 2.9 };
        if (l1a_products[i_l1a].get().exposure_time < exposure_thres) {
            continue;
        }
        for (int i {}; i < ckd.npix; ++i) {
            values_for_median[i] = l1a_products[i_l1a].get().image[i];
        }
        const double med { median(values_for_median) };
        if (med < clipping_threshold) {
            I_ref.push_back(l1a_products[i_l1a].get().meta->reference_signal);
            ckd.nonlin.S_max_data.push_back(med);
        }
    }
    // Medians of linear detector currents with dark current
    ckd.nonlin.I_c_med.resize(I_ref.size());
    for (int i_lev {}; i_lev < static_cast<int>(I_ref.size()); ++i_lev) {
        for (int i {}; i < ckd.npix; ++i) {
            values_for_median[i] =
              ckd.dark.current[i] + ckd.nonlin.f1[i] * I_ref[i_lev]
              + ckd.nonlin.f2[i]
                  * std::pow(ckd.nonlin.f1[i] * I_ref[i_lev], 1.5);
        }
        ckd.nonlin.I_c_med[i_lev] = median(values_for_median);
    }
    gadfit::LMsolver solver { [](const std::vector<gadfit::AdVar>& parameters,
                                 const double x) -> gadfit::AdVar {
        return parameters[0] + parameters[1] * std::log(x)
               + parameters[2] / std::sqrt(x);
    } };
    solver.addDataset(ckd.nonlin.I_c_med, ckd.nonlin.S_max_data);
    solver.setPar(0, 1e3, true);
    solver.setPar(1, 1e1, true);
    solver.setPar(2, -1e1, true);
    solver.settings.verbosity = gadfit::io::final_only;
    solver.fit();
    ckd.nonlin.m0 = { true, mpi, ckd.npix };
    if (mpi.rank == 0) {
        std::ranges::fill(ckd.nonlin.m0, solver.getParValue(0));
    }
    ckd.nonlin.m1 = solver.getParValue(1);
    ckd.nonlin.m2 = solver.getParValue(2);
    // Sort for easier visualization (these are not used anymore)
    std::ranges::sort(ckd.nonlin.I_c_med);
    std::ranges::sort(ckd.nonlin.S_max_data);
    for (const auto& I_c : ckd.nonlin.I_c_med) {
        ckd.nonlin.S_max_fit.push_back(solver.getValue(I_c));
    }
}

auto NonlinGen::fitSigmoidal(
  const std::vector<std::reference_wrapper<L1A>>& l1a_products) const -> void
{
    resize(ckd.nonlin.r, mpi, ckd.nonlin_n_r, ckd.npix);
    ckd.nonlin.c = { true, mpi, ckd.npix };
    // Arrays to hold the linear and measured signals. Number of
    // points can be different for each pixel.
    std::vector<double> S_lin {};
    std::vector<double> S_meas {};
    // Like S_max but without m0: S_max_partial = m1 * ln(Ic) + m2 / sqrt(Ic)
    std::vector<double> S_max_partial {};
    // The index array serves as x-values for the solver. Its values
    // refers to elements in the S_lin and S_max_partial arrays inside
    // the solver.
    std::vector<double> S_idx {};
    // If debug pixels were requested, set up additional arrays to be
    // saved in the CKD
    ckd.nonlin.debug_pixels = settings.nonlin.debug_pixels;
    const int npix_d { static_cast<int>(ckd.nonlin.debug_pixels.size()) };
    if (settings.debug.at(static_cast<int>(ProcLevel::nonlin)).value()
        && !ckd.nonlin.debug_pixels.empty()) {
        const int n_max_signals { static_cast<int>(l1a_products.size()) };
        resize(ckd.nonlin.scaled_S_lin, mpi, npix_d, n_max_signals);
        resize(ckd.nonlin.scaled_S_meas, mpi, npix_d, n_max_signals);
        resize(ckd.nonlin.sigmoid_fits, mpi, npix_d, n_max_signals);
        for (int i_pix {}; i_pix < npix_d; ++i_pix) {
            std::ranges::fill(ckd.nonlin.scaled_S_lin.at(i_pix), 0.0);
            std::ranges::fill(ckd.nonlin.scaled_S_meas.at(i_pix), 0.0);
            std::ranges::fill(ckd.nonlin.sigmoid_fits.at(i_pix), 0.0);
        }
    }
    // Define the fitting function (passed to the solver later)
    const auto sigmoidal { [&S_lin, &S_max_partial](
                             const std::vector<gadfit::AdVar>& pars,
                             const double idx_d) -> gadfit::AdVar {
        using gadfit::AdVar;
        const int idx { static_cast<int>(idx_d) };
        const AdVar S_max { pars[6] + S_max_partial[idx] };
        const AdVar x { S_lin[idx] / S_max };
        const AdVar x2 { x * x };
        constexpr double eps { 1e-10 };
        const AdVar ln_x { log(abs(x) + eps) };
        const AdVar r { pars[0] + pars[1] * ln_x + pars[2] * x2
                        + pars[3] * x2 * ln_x + pars[4] * x2 * abs(x) };
        const AdVar& c { pars[5] };
        const AdVar xrc { x * r * c / tanh(c) };
        return S_max / (2.0 * c) * log(cosh(c + xrc) / cosh(c - xrc));
    } };
    // For printing on screen
    const int percentage_factor { pix_ranges.at(mpi.rank + 1) / 1000 };
    // Main loop that is parallelized over pixels
    for (int i_pix { pix_ranges.at(mpi.rank) };
         i_pix < pix_ranges.at(mpi.rank + 1);
         ++i_pix) {
        if (percentage_factor == 0 || i_pix % percentage_factor == 0) {
            spdlog::info(
              "Fitting the sigmoidal function {:5.1f}%",
              1e2 * i_pix
                / (pix_ranges.at(mpi.rank + 1) - pix_ranges.at(mpi.rank)));
        }
        // These arrays can have different lengths for each pixel
        S_lin.resize(0);
        S_meas.resize(0);
        S_max_partial.resize(0);
        S_idx.resize(0);
        const double clipping_threshold {
            1020.0 - l1a_products.front().get().digital_offset_single
            - ckd.dark.offset_long[i_pix]
        };
        constexpr double exposure_thres { 1700.0 };
        constexpr double current_thres { 50.0e3 }; // per ms
        for (const auto& l1a : l1a_products) {
            const double I_lin { ckd.dark.current[i_pix]
                                 + ckd.nonlin.f1[i_pix]
                                     * l1a.get().meta->reference_signal };
            const double I_c_nlin { std::max(I_lin - ckd.dark.current[i_pix],
                                             0.0) };
            const double I_c {
                I_lin + ckd.nonlin.f2[i_pix] * I_c_nlin * std::sqrt(I_c_nlin)
            };
            if (l1a.get().image[i_pix] < clipping_threshold
                && I_lin * l1a.get().exposure_time < exposure_thres
                && I_lin < current_thres) {
                S_lin.push_back(I_c * l1a.get().exposure_time);
                S_meas.push_back(l1a.get().image[i_pix]);
                S_max_partial.push_back(ckd.nonlin.m1 * std::log(I_c)
                                        + ckd.nonlin.m2 / std::sqrt(I_c));
            }
        }
        S_idx.resize(S_lin.size());
        std::iota(S_idx.begin(), S_idx.end(), 0.0);
        // Set up solver
        gadfit::LMsolver solver { sigmoidal };
        solver.addDataset(S_idx, S_meas);
        solver.settings.iteration_limit = 100;
        solver.settings.loss = gadfit::Loss::cauchy;
        // Set initial parameters
        solver.setPar(0, 1.00, true);                 // r0
        solver.setPar(1, 0.020, true);                // r1
        solver.setPar(2, -0.01, true);                // r2
        solver.setPar(3, -0.01, true);                // r3
        solver.setPar(4, 0.01, true);                 // r4
        solver.setPar(5, 7.0, true);                  // c
        solver.setPar(6, ckd.nonlin.m0[i_pix], true); // m0
        // Perform fit
        loggerOnOff();
        solver.fit();
        loggerOnOff();
        // Retrieve parameters
        for (int i_r {}; i_r < ckd.nonlin_n_r; ++i_r) {
            ckd.nonlin.r[i_r][i_pix] = solver.getParValue(i_r);
        }
        ckd.nonlin.c[i_pix] = solver.getParValue(5);
        ckd.nonlin.m0[i_pix] = solver.getParValue(6);
        // Save debugging information
        if (!ckd.nonlin.sigmoid_fits.empty()) {
            for (int i_pix_d {}; i_pix_d < npix_d; ++i_pix_d) {
                if (ckd.nonlin.debug_pixels[i_pix_d] == i_pix) {
                    for (int i_lin {}; i_lin < static_cast<int>(S_lin.size());
                         ++i_lin) {
                        const double S_max { ckd.nonlin.m0[i_pix]
                                             + S_max_partial[i_lin] };
                        ckd.nonlin.scaled_S_lin[i_pix_d][i_lin] =
                          S_lin[i_lin] / S_max;
                        ckd.nonlin.scaled_S_meas[i_pix_d][i_lin] =
                          S_meas[i_lin] / S_max;
                        ckd.nonlin.sigmoid_fits[i_pix_d][i_lin] =
                          solver.getValue(static_cast<double>(i_lin)) / S_max;
                    }
                    break;
                }
            }
        }
    }
}

// Here is the sigmoidal function again but rewritten for faster
// evaluation. The one used in the solver is much slower.
[[nodiscard]] static auto sigmoidal(const CKD& ckd,
                                    const double S_lin,
                                    const double exposure_time,
                                    const int i_pix) -> double
{
    const double I_c { S_lin / exposure_time };
    const double S_max { ckd.nonlin.m0[i_pix] + ckd.nonlin.m1 * std::log(I_c)
                         + ckd.nonlin.m2 / std::sqrt(I_c) };
    const double x { S_lin / S_max };
    const double x2 { x * x };
    constexpr double eps { 1e-10 };
    const double ln_x { std::log(std::abs(x) + eps) };
    const double r { ckd.nonlin.r[0][i_pix] + ckd.nonlin.r[1][i_pix] * ln_x
                     + ckd.nonlin.r[2][i_pix] * x2
                     + ckd.nonlin.r[3][i_pix] * x2 * ln_x
                     + ckd.nonlin.r[4][i_pix] * x2 * std::abs(x) };
    const double c { ckd.nonlin.c[i_pix] };
    const double xrc { x * r * c / std::tanh(c) };
    return S_max / (2.0 * c)
           * std::log(std::cosh(c + xrc) / std::cosh(c - xrc));
}

auto NonlinGen::parametrizeSigmoidal() const -> void
{
    // Interpolation is first done on a dense (S_lin, S_meas) grid and
    // then interpolated onto a sparser (S_meas, S_lin) grid.
    constexpr int knot_multiplier { 10 };
    const int n_S_lin_dense { knot_multiplier * settings.nonlin.n_knots };
    std::vector<double> S_lin_dense(n_S_lin_dense);
    for (int i {}; i < n_S_lin_dense; ++i) {
        constexpr double c_min { 1e-10 };
        const double& c_max { settings.nonlin.target_counts_max };
        S_lin_dense[i] = c_min + i * (c_max - c_min) / (n_S_lin_dense - 1);
    }
    // Define knots for the CKD nonlinearity spline. These are the
    // same for each pixel and won't change. Note that the dense knot
    // vector (S_lin_dense) refers to the linear signals. The CKD knot
    // vector refers to measured signals and its largest value is
    // determined by the sigmoidal function using the last value of
    // S_lin_dense.
    ckd.nonlin.spline = {
        settings.nonlin.spline_type, mpi, ckd.npix, settings.nonlin.n_knots
    };
    if (mpi.rank == 0) {
        const double c_min { static_cast<double>(
          settings.nonlin.target_counts_min) };
        const double c_max { sigmoidal(ckd,
                                       S_lin_dense.back(),
                                       settings.nonlin.target_exposure_time,
                                       ckd.npix / 2) };
        for (int i {}; i < settings.nonlin.n_knots; ++i) {
            ckd.nonlin.spline.knots[i] =
              c_min + i * (c_max - c_min) / (settings.nonlin.n_knots - 1);
        }
    }
#ifdef USE_MPI
    MPI_Barrier(mpi.comm);
#endif
    std::vector<double> S_meas_dense(n_S_lin_dense);
    std::vector<double> S_lin_sparse(settings.nonlin.n_knots);
    const int percentage_factor { pix_ranges.at(mpi.rank + 1) / 100 };
    for (int i_pix { pix_ranges.at(mpi.rank) };
         i_pix < pix_ranges.at(mpi.rank + 1);
         ++i_pix) {
        if (percentage_factor == 0 || i_pix % percentage_factor == 0) {
            spdlog::info(
              "Parameterizing the fit with splines {:5.1f}%",
              1e2 * i_pix
                / (pix_ranges.at(mpi.rank + 1) - pix_ranges.at(mpi.rank)));
        }
        for (int i_knot {}; i_knot < n_S_lin_dense; ++i_knot) {
            S_meas_dense[i_knot] =
              sigmoidal(ckd,
                        S_lin_dense[i_knot],
                        settings.nonlin.target_exposure_time,
                        i_pix);
        }
        // Now that we have the dense mapping, interpolate the linear
        // signal values using the CKD knot vector (measured signals)
        // which is shared by all pixels.
        if (settings.nonlin.spline_type == SplineType::linear) {
            LinearSpline spline { S_meas_dense, S_lin_dense };
            for (int i {}; i < settings.nonlin.n_knots; ++i) {
                S_lin_sparse[i] = spline.eval(ckd.nonlin.spline.knots[i]);
            }
        } else {
            CubicSpline spline { S_meas_dense, S_lin_dense };
            for (int i {}; i < settings.nonlin.n_knots; ++i) {
                S_lin_sparse[i] = spline.eval(ckd.nonlin.spline.knots[i]);
            }
        }
        ckd.nonlin.spline.generateSpline(i_pix, S_lin_sparse);
    }
}

static auto readInputCKD(const std::string filename, CKD& ckd) -> void
{
    NcObject ckd_in { netCDF::NcFile::read, filename };
    const auto grp { ckd_in.getGroup("DEBUG").getGroup("NON_LINEARITY") };
    ckd.readNonlinDebugData(grp);
}

auto NonlinGen::processBatchSigmoidal() -> void
{
    if (settings.nonlin.restart_ckd.empty()) {
        // We start with dark offset and current corrected data. After
        // the first step the signal again includes dark current.
        addSignalInduced(settings.nonlin.signal_induced_offset,
                         batches.front().l1a_products);
        computeSignalRates(batches.front().l1a_products);
        computeRefCurrent();
        determineMaximumSignals(batches.front().l1a_products);
        fitSigmoidal(batches.front().l1a_products);
        // The L1A products are not needed anymore. Free memory before
        // starting with the parameterization.
        batches.front().destroy();
    } else {
        batches.front().destroy();
        readInputCKD(settings.nonlin.restart_ckd, ckd);
    }
    parametrizeSigmoidal();
}

auto NonlinGen::processMedian() -> void
{
    const auto l1a { std::ranges::max(
      batches.front().l1a_products, [](const auto& a, const auto& b) {
          return a.get().meta->illumination_level
                 < b.get().meta->illumination_level;
      }) };
    if (settings.nonlin.scale_median) {
        double med {};
        median(ckd.npix, l1a.get().image.data(), med);
        for (auto& val : ckd.nonlin_signal_scale) {
            val = med;
        }
    } else {
        for (int i {}; i < ckd.npix; ++i) {
            ckd.nonlin_signal_scale[i] = l1a.get().image[i];
        }
    }
}

auto NonlinGen::processLin() -> void
{
    const auto& l1a_products { batches.front().l1a_products };
    // Linear result pointers
    double* linslopeptr = &nonlin_lin_slope[0];
    double* linchi2ptr = &nonlin_lin_chi2[0];

    spdlog::info("Calculating linear fit per pixel");
    for (size_t ipix {}; ipix < ckd.npix; ipix++) {
        // Put fillvalues everywhere
        *linslopeptr = NC_FILL_DOUBLE;
        *linchi2ptr = NC_FILL_DOUBLE;

        // Breakable do-once loop
        for (bool ctd = true; ctd; ctd = false) {

            // Die if you are already dead
            if (ckd.pixel_mask[ipix])
                break;
            ckd.pixel_mask[ipix] =
              true; // Unless you survive until the end of the loop

            // Gather the measurements at this pixel index
            std::vector<double> signal(l1a_products.size());
            std::vector<double> noise(l1a_products.size());
            std::vector<bool> pixel_mask(l1a_products.size());
            for (size_t il1a = 0; il1a < l1a_products.size(); il1a++) {
                signal[il1a] = l1a_products[il1a].get().image[ipix];
                noise[il1a] = l1a_products[il1a].get().noise[ipix];
                pixel_mask[il1a] = l1a_products[il1a].get().pixel_mask[ipix];
            }
            // Filter for signals between the chosen range
            size_t nl1a_lin = 0;
            size_t nl1a_lin_healthy = 0;
            // Maximum allocation size
            std::vector<double> abscissa_lin(l1a_products.size());
            std::vector<double> signal_lin(l1a_products.size());
            std::vector<double> s_y_lin(l1a_products.size());
            std::vector<bool> pixel_mask_lin(l1a_products.size());
            for (size_t il1a = 0; il1a < l1a_products.size(); il1a++) {
                double& sig = signal[il1a];
                if (sig > settings.nonlin.lmin && sig < settings.nonlin.lmax) {
                    abscissa_lin[nl1a_lin] = abscissa[il1a];
                    signal_lin[nl1a_lin] = sig;
                    s_y_lin[nl1a_lin] = pow(noise[il1a], 2.0);
                    pixel_mask_lin[nl1a_lin] = pixel_mask[il1a];
                    if (!pixel_mask_lin[nl1a_lin]) {
                        nl1a_lin_healthy++;
                    }
                    ++nl1a_lin;
                }
            }

            // Encapsulate vectors to their intended sizes
            abscissa_lin.resize(nl1a_lin);
            signal_lin.resize(nl1a_lin);
            s_y_lin.resize(nl1a_lin);
            pixel_mask_lin.resize(nl1a_lin);

            // Nothing is possible if there are no reliable measurments
            if (nl1a_lin_healthy < 1)
                break;

            // Perform the linear fit. Note that if it fails, no mess is created
            // in the result before returning.
            if (linear_invert(1,
                              nl1a_lin,
                              abscissa_lin.data(),
                              OPT_DIAG,
                              s_y_lin.data(),
                              &pixel_mask_lin,
                              signal_lin.data(),
                              linslopeptr)
                != 0)
                break;
            *linchi2ptr = 0.0;
            for (size_t il1a = 0; il1a < nl1a_lin; il1a++) {
                if (pixel_mask_lin[il1a])
                    continue;
                *linchi2ptr +=
                  pow((abscissa_lin[il1a] * *linslopeptr - signal_lin[il1a]),
                      2.0)
                  / s_y_lin[il1a] / (nl1a_lin_healthy - 1);
            }

            // Apply mask criterion: Linear chi squared
            if (settings.nonlin.mask_lin_chi2_max != NC_FILL_DOUBLE
                && *linchi2ptr > settings.nonlin.mask_lin_chi2_max)
                break;

            // Survived to the end of the loop
            ckd.pixel_mask[ipix] =
              false; // Unless you survive until the end of the loop
        }

        // Progress pointers
        linslopeptr++;
        linchi2ptr++;
    }
}

auto NonlinGen::processNonlin() -> void
{
    const auto& l1a_products { batches.front().l1a_products };
    // Pointers to earlier calculated results
    double* linslopeptr = &nonlin_lin_slope[0];
    double* scaleptr = &ckd.nonlin_signal_scale[0];
    // Pointers to results
    double* resptr = &ckd.nonlin_fit[0];
    double* chi2ptr = &nonlin_chi2[0];

    spdlog::info("Calculating nonlinearity fit per pixel");
    for (size_t ipix {}; ipix < ckd.npix; ipix++) {
        // Fill fillvalues everywhere
        for (size_t ispline = 0; ispline < ckd.dim_nonlin_spline; ispline++)
            resptr[ispline] = NC_FILL_DOUBLE;
        *chi2ptr = NC_FILL_DOUBLE;

        for (bool ctd = true; ctd; ctd = false) {
            if (ckd.pixel_mask[ipix]) {
                break;
            }
            ckd.pixel_mask[ipix] =
              true; // Unless you survive until the end of the loop

            std::vector<double> signal_corrected_scaled(l1a_products.size());
            std::vector<double> signal_measured(l1a_products.size());
            std::vector<double> s_y_measured(l1a_products.size());
            std::vector<bool> pixel_mask_measured(l1a_products.size());
            size_t nl1a_healthy = 0;
            for (size_t il1a = 0; il1a < l1a_products.size(); il1a++) {
                L1A* l1a = &l1a_manager->l1a_products[il1a];
                signal_corrected_scaled[il1a] =
                  *linslopeptr * l1a->exposure_time
                  * l1a->meta->illumination_level / (*scaleptr);
                signal_measured[il1a] = l1a->image[ipix];
                s_y_measured[il1a] = pow(l1a->noise[ipix], 2.0);
                pixel_mask_measured[il1a] = l1a->pixel_mask[ipix];
                if (!pixel_mask_measured[il1a])
                    nl1a_healthy++;
            }

            // Perform the fit
            std::vector<double> jac(ckd.dim_nonlin_spline * l1a_products.size(),
                                    0.0);
            b->jaccalc(
              l1a_products.size(), signal_corrected_scaled.data(), jac.data());
            // Add hard constraint at zero. Zero signal should mean zero
            // mismatch.
            double zero = 0.0; // Used as abscissa value and as measurement
            std::vector<double> jac_zero(ckd.dim_nonlin_spline);
            b->jaccalc(1, &zero, jac_zero.data());
            if (constrained_linear_invert(ckd.dim_nonlin_spline,
                                          l1a_products.size(),
                                          1,
                                          jac.data(),
                                          jac_zero.data(),
                                          OPT_DIAG,
                                          s_y_measured.data(),
                                          &pixel_mask_measured,
                                          signal_measured.data(),
                                          &zero,
                                          resptr)
                != 0)
                break;
            *chi2ptr = 0.0;
            std::vector<double> mod(l1a_products.size());
            Matrix::matmul_rl_fold_slow(ckd.dim_nonlin_spline,
                                        l1a_products.size(),
                                        jac.data(),
                                        resptr,
                                        mod.data());
            for (size_t il1a = 0; il1a < l1a_products.size(); il1a++) {
                if (pixel_mask_measured[il1a])
                    continue;
                // The plus one is the hard constraint, reducing the
                // DFS by one (or adding a measurement with no bias).
                *chi2ptr += pow(mod[il1a] - signal_measured[il1a], 2.0)
                            / s_y_measured[il1a]
                            / (nl1a_healthy - ckd.dim_nonlin_spline + 1);
            }

            // Apply mask criterion: Nonlinear chi squared
            if (settings.nonlin.mask_chi2_max
                && *chi2ptr > settings.nonlin.mask_chi2_max)
                break;

            // Survived to the end of the loop
            ckd.pixel_mask[ipix] = false;
        }

        // Progress pointers
        linslopeptr++;
        scaleptr += 1;
        resptr += ckd.dim_nonlin_spline;
        chi2ptr += 1;
    }
}

auto NonlinGen::processBatchSpline() -> void
{
    // Apply signal-induced offset
    addSignalInduced(settings.nonlin.signal_induced_offset,
                     batches.front().l1a_products);
    // Subtract dark current
    for (auto& l1a : batches.front().l1a_products) {
        for (int i {}; i < ckd.npix; ++i) {
            l1a.get().image[i] -= ckd.dark.current[i] * l1a.get().exposure_time;
        }
    }
    processMedian();
    processLin();
    processNonlin();
}

auto NonlinGen::processBatch(const int /* i_batch */) -> void
{
    if (settings.nonlin.model == NonlinModel::sigmoidal) {
        processBatchSigmoidal();
    } else {
        processBatchSpline();
    }
}

} // namespace spex
