// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for generating the nonlinearity CKD. The equation numbers
// refer to Paul's tech note SRON-SPEX-TN-2020-001 issue 0.5.

#pragma once

#include "processor.h"

#include <vector>

namespace spex {

class Bspline;

class NonlinGen : public Processor
{
private:
    // Used when parallelizing over pixels
    std::vector<int> pix_ranges {};
    // Slope fit during linear fit
    std::vector<double> nonlin_lin_slope {};
    // Residual chi squared on the linear fit
    std::vector<double> nonlin_lin_chi2 {};
    // Residual chi squared on non-linearity correction
    std::vector<double> nonlin_chi2 {};
    // Exposure time times illumination level
    std::vector<double> abscissa {};
    // Fitting spline
    std::unique_ptr<Bspline> b {};
    std::vector<int> nonlin_il1a_tracker {};

    auto processInit() -> void override;

    // Modify signal by adding signal-induced offset to it
    // [Eq. (80)]. After this the image contains the dark current. If
    // !active do not add signal induced offset, only add dark
    // current.
    auto addSignalInduced(
      const bool active,
      std::vector<std::reference_wrapper<L1A>>& l1a_products) const -> void;
    // Compute signal rates per light intensity (step 4)
    auto computeSignalRates(
      const std::vector<std::reference_wrapper<L1A>>& l1a_products) const
      -> void;
    // Determine average reference current (step 5)
    auto computeRefCurrent() const -> void;
    // Determine offset corrected maximum signals from detector medians (steps 7
    // and 8)
    auto determineMaximumSignals(
      const std::vector<std::reference_wrapper<L1A>>& l1a_products) const
      -> void;
    // Fit the nonlinearity model (step 9, Eqs. (72) and (73))
    auto fitSigmoidal(
      const std::vector<std::reference_wrapper<L1A>>& l1a_products) const
      -> void;
    // Compute the (S_meas, S_lin) curves for each pixel and
    // parametrize them with cubic splines
    auto parametrizeSigmoidal() const -> void;
    // Generate the nonlinearity CKD using the sigmoidal function model
    auto processBatchSigmoidal() -> void;

    // Previous implementation based on B-splines
    auto processMedian() -> void;
    auto processLin() -> void;
    auto processNonlin() -> void;
    auto processBatchSpline() -> void;

    // Driver routine that selects the model to run
    auto processBatch(const int /* ibatch */) -> void override;

public:
    NonlinGen(const gadfit::MPIVars& mpi,
              const SPEXSettings& settings,
              CKD& ckd)
      : Processor { mpi, settings, ProcLevel::nonlin, ckd }
    {}
    ~NonlinGen() = default;
};

} // namespace spex
