// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "spline_multi.h"

#include "cubic_spline.h"
#include "spex_math.h"

#include <gadfit/mpi.h>

namespace spex {

SplineMulti::SplineMulti(const SplineType type,
                         const gadfit::MPIVars& mpi,
                         const int n_curves,
                         const int n_knots)
  : type { type }
  , n_curves { n_curves }
  , n_knots { n_knots }
  , n_knots_1 { n_knots - 1 }
{
    knots = { true, mpi, n_knots };
    y = { true, mpi, n_curves * n_knots };
    if (type == SplineType::cubic) {
        B = { true, mpi, n_curves * n_knots };
        A = { true, mpi, n_curves * n_knots_1 };
        C = { true, mpi, n_curves * n_knots_1 };
    }
}

auto SplineMulti::setLookupIdxScale() -> void
{
    lookup_idx_scale =
      static_cast<int>(knots.size() - 1) / (knots.back() - knots.front());
}

auto SplineMulti::generateSpline(const int i_curve,
                                 const std::vector<double>& values) -> void
{
    setLookupIdxScale();
    for (int i_knot {}; i_knot < n_knots; ++i_knot) {
        const int idx { i_curve * n_knots + i_knot };
        y[idx] = values[i_knot];
    }
    if (type == SplineType::cubic) {
        std::vector<double> knots_v(n_knots);
        for (int i {}; i < n_knots; ++i) {
            knots_v[i] = knots[i];
        }
        CubicSpline spline { knots_v, values };
        for (int i_knot {}; i_knot < n_knots; ++i_knot) {
            const int idx { i_curve * n_knots + i_knot };
            B[idx] = spline.getB()[i_knot];
        }
        for (int i_knot {}; i_knot < n_knots_1; ++i_knot) {
            const int idx { i_curve * n_knots_1 + i_knot };
            A[idx] = spline.getA()[i_knot];
            C[idx] = spline.getC()[i_knot];
        }
    }
}

auto SplineMulti::lookupIdx(const double x) const -> int
{
    return static_cast<int>(lookup_idx_scale * (x - knots.front()));
}

[[nodiscard]] auto SplineMulti::eval(const int i_curve, const double x) const
  -> double
{
    const int idx { equal_spacing ? lookupIdx(x) : binaryFindIdx(knots, x) };
    if (type == SplineType::linear) {
        // Duplicate first or last knots outside the interpolation range
        if (x <= knots.front()) {
            const double t { (x - knots.front()) / (knots[1] - knots.front()) };
            return std::lerp(y[i_curve * n_knots], y[i_curve * n_knots + 1], t);
        }
        if (x >= knots.back()) {
            const double& knots_2 { knots[n_knots - 2] };
            const double t { (x - knots_2) / (knots.back() - knots_2) };
            return std::lerp(y[(i_curve + 1) * n_knots - 2],
                             y[(i_curve + 1) * n_knots - 1],
                             t);
        }
        const double t { (x - knots[idx]) / (knots[idx + 1] - knots[idx]) };
        return std::lerp(
          y[i_curve * n_knots + idx], y[i_curve * n_knots + idx + 1], t);
    } // else cubic
    // Do linear extrapolation outside the interpolation range
    if (x <= knots.front()) {
        return y[i_curve * n_knots]
               + A[i_curve * n_knots_1] * (x - knots.front());
    }
    if (x >= knots.back()) {
        return y[(i_curve + 1) * n_knots - 1]
               + A[(i_curve + 1) * n_knots_1 - 1] * (x - knots.back());
    }
    const double Dx { x - knots[idx] };
    const double Dx2 { Dx * Dx };
    const double Dx3 { Dx2 * Dx };
    return y[i_curve * n_knots + idx] + A[i_curve * n_knots_1 + idx] * Dx
           + B[i_curve * n_knots + idx] * Dx2
           + C[i_curve * n_knots_1 + idx] * Dx3;
}

[[nodiscard]] auto SplineMulti::deriv(const int i_curve, const double x) const
  -> double
{
    const int idx { equal_spacing ? lookupIdx(x) : binaryFindIdx(knots, x) };
    if (type == SplineType::linear) {
        // Duplicate first or last derivatives outside the interpolation range
        if (x <= knots.front()) {
            return (y[i_curve * n_knots + 1] - y[i_curve * n_knots])
                   / (knots[1] - knots.front());
        }
        if (x >= knots.back()) {
            return (y[(i_curve + 1) * n_knots - 1]
                    - y[(i_curve + 1) * n_knots - 2])
                   / (knots.back() - knots[n_knots - 2]);
        }
        return (y[i_curve * n_knots + idx + 1] - y[i_curve * n_knots + idx])
               / (knots[idx + 1] - knots[idx]);
    } // else cubic
    // Do linear extrapolation outside the interpolation range
    if (x <= knots.front()) {
        return A[i_curve * n_knots_1];
    }
    if (x >= knots.back()) {
        return A[(i_curve + 1) * n_knots_1 - 1];
    }
    const double Dx { x - knots[idx] };
    const double Dx2 { Dx * Dx };
    return A[i_curve * n_knots_1 + idx] + 2 * B[i_curve * n_knots + idx] * Dx
           + 3 * C[i_curve * n_knots_1 + idx] * Dx2;
}

} // namespace spex
